##################### generated by xml-casa (v2) from subvs.xml #####################
##################### 359882fd4602fc4cb310cc193180fe92 ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from .private.task_subvs import subvs as _subvs_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log

class _subvs:
    """
    subvs ---- Vector-subtraction in UV using selected time ranges and spectral channels as background

    
    
    
    Split is the general purpose program to make a new data set that is a
    subset or averaged form of an existing data set.  General selection
    parameters are included, and one or all of the various data columns
    (DATA, LAG_DATA and/or FLOAT_DATA, and possibly MODEL_DATA and/or
    CORRECTED_DATA) can be selected.
    
    Split is often used after the initial calibration of the data to make a
    smaller measurement set with only the data that will be used in
    further flagging, imaging and/or self-calibration.  split can
    average over frequency (channels) and time (integrations).

    --------- parameter descriptions ---------------------------------------------

    vis         Name of input measurement set
    outputvis   Name of output measurement set
    timerange   Select the time range of the input visbility to be subtracted from
    spw         Select the spectral channels of the input visibility to be subtracted from
    mode        Operation: linear, highpass
    subtime1    Select the first time range as the background for uv subtraction
    subtime2    Select the second time range as the background for uv subtraction
    smoothaxis  Select the axis along which smooth is performed
    smoothtype  Select the smooth type
    smoothwidth Select the width of the smoothing window
    splitsel    Split the selected timerange and spectral channels as outputvis
    reverse     Reverse the sign of the background-subtracted data (for absorptive structure)
    overwrite   Overwrite the already existing output measurement set

    --------- examples -----------------------------------------------------------

    
    
    Subvs is a task to do UV vector-subtraction, by selecting time ranges
    in the data as background. Subvs can be used to subtract the background
    continuum emission to separate the time-dependent emission, e.g. solar
    coherent radio bursts.
    
    Keyword arguments:
    vis -- Name of input visibility file (MS)
    default: none; example: vis='ngc5921.ms'
    outputvis -- Name of output uv-subtracted visibility file (MS)
    default: none; example: outputvis='ngc5921_src.ms'
    timerange -- Time range of performing the UV subtraction:
    default='' means all times.  examples:
    timerange = 'YYYY/MM/DD/hh:mm:ss~YYYY/MM/DD/hh:mm:ss'
    timerange = 'hh:mm:ss~hh:mm:ss'
    spw -- Select spectral window/channel.
    default = '' all the spectral channels. Example: spw='0:1~20'
    mode -- operation mode
    default 'linear'
    mode = 'linear': use a linear fit for the background to be subtracted
    mode = 'lowpass': act as a lowpass filter---smooth the data using different smooth
    types and smooth window size. Can be performed along either time
    or frequency axis
    mode = 'highpass': act as a highpass filter---smooth the data first, and
    subtract the smoothed data from the original. Can be performed along either time
    or frequency axis
    mode = 'linear' expandable parameters:
    subtime1 -- Time range 1 of the background to be subtracted from the data
    default='' means all times.  format:
    timerange = 'YYYY/MM/DD/hh:mm:ss~YYYY/MM/DD/hh:mm:ss'
    timerange = 'hh:mm:ss~hh:mm:ss'
    subtime2 -- Time range 2 of the backgroud to be subtracted from the data
    default='' means all times.  examples:
    timerange = 'YYYY/MM/DD/hh:mm:ss~YYYY/MM/DD/hh:mm:ss'
    timerange = 'hh:mm:ss~hh:mm:ss'
    mode = 'lowpass' or 'highpass' expandable parameters:
    smoothaxis -- axis of smooth
    Default: 'time'
    smoothaxis = 'time': smooth is along the time axis
    smoothaxis = 'freq': smooth is along the frequency axis
    smoothtype -- type of the smooth depending on the convolving kernel
    Default: 'flat'
    smoothtype = 'flat': convolving kernel is a flat rectangle,
    equivalent to a boxcar moving smooth
    smoothtype = 'hanning': Hanning smooth kernel. See numpy.hanning
    smoothtype = 'hamming': Hamming smooth kernel. See numpy.hamming
    smoothtype = 'bartlett': Bartlett smooth kernel. See numpy.bartlett
    smoothtype = 'blackman': Blackman smooth kernel. See numpy.blackman
    smoothwidth -- width of the smooth kernel
    Default: 5
    Examples: smoothwidth=5, meaning the width is 5 pixels
    splitsel -- True or False. default = False. If splitsel = False, then the entire input
    measurement set is copied as the output measurement set (outputvis), with
    background subtracted at selected timerange and spectral channels.
    If splitsel = True,then only the selected timerange and spectral channels
    are copied into the output measurement set (outputvis).
    reverse -- True or False. default = False. If reverse = False, then the times indicated
    by subtime1 and/or subtime2 are treated as background and subtracted; If reverse
    = True, then reverse the sign of the background-subtracted data. The option can
    be used for mapping absorptive structure.
    overwrite -- True or False. default = False. If overwrite = True and
    outputvis already exists, the selected subtime and spw in the
    output measurment set will be replaced with background subtracted
    visibilities


    """

    _info_group_ = """misc"""
    _info_desc_ = """Vector-subtraction in UV using selected time ranges and spectral channels as background"""

    def __call__( self, vis='', outputvis='', timerange='', spw='', mode='linear', subtime1='', subtime2='', smoothaxis='time', smoothtype='flat', smoothwidth=int(5), splitsel=True, reverse=False, overwrite=False ):
        schema = {'vis': {'type': 'cReqPath', 'coerce': _coerce.expand_path}, 'outputvis': {'type': 'cPath', 'coerce': _coerce.expand_path}, 'timerange': {'type': 'cStr', 'coerce': _coerce.to_str}, 'spw': {'type': 'cStr', 'coerce': _coerce.to_str}, 'mode': {'type': 'cStr', 'coerce': _coerce.to_str, 'allowed': [ 'linear', 'lowpass', 'highpass' ]}, 'subtime1': {'type': 'cStr', 'coerce': _coerce.to_str}, 'subtime2': {'type': 'cStr', 'coerce': _coerce.to_str}, 'smoothaxis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'smoothtype': {'type': 'cStr', 'coerce': _coerce.to_str}, 'smoothwidth': {'type': 'cInt'}, 'splitsel': {'type': 'cBool'}, 'reverse': {'type': 'cBool'}, 'overwrite': {'type': 'cBool'}}
        doc = {'vis': vis, 'outputvis': outputvis, 'timerange': timerange, 'spw': spw, 'mode': mode, 'subtime1': subtime1, 'subtime2': subtime2, 'smoothaxis': smoothaxis, 'smoothtype': smoothtype, 'smoothwidth': smoothwidth, 'splitsel': splitsel, 'reverse': reverse, 'overwrite': overwrite}
        assert _pc.validate(doc,schema), str(_pc.errors)
        _logging_state_ = _start_log( 'subvs', [ 'vis=' + repr(_pc.document['vis']), 'outputvis=' + repr(_pc.document['outputvis']), 'timerange=' + repr(_pc.document['timerange']), 'spw=' + repr(_pc.document['spw']), 'mode=' + repr(_pc.document['mode']), 'subtime1=' + repr(_pc.document['subtime1']), 'subtime2=' + repr(_pc.document['subtime2']), 'smoothaxis=' + repr(_pc.document['smoothaxis']), 'smoothtype=' + repr(_pc.document['smoothtype']), 'smoothwidth=' + repr(_pc.document['smoothwidth']), 'splitsel=' + repr(_pc.document['splitsel']), 'reverse=' + repr(_pc.document['reverse']), 'overwrite=' + repr(_pc.document['overwrite']) ] )
        return _end_log( _logging_state_, 'subvs', _subvs_t( _pc.document['vis'], _pc.document['outputvis'], _pc.document['timerange'], _pc.document['spw'], _pc.document['mode'], _pc.document['subtime1'], _pc.document['subtime2'], _pc.document['smoothaxis'], _pc.document['smoothtype'], _pc.document['smoothwidth'], _pc.document['splitsel'], _pc.document['reverse'], _pc.document['overwrite'] ) )

subvs = _subvs( )




import argparse
from datetime import datetime, timedelta
from astropy.time import Time
from suncasa.suncasatasks import ptclean6 as ptclean
from suncasa.suncasatasks import calibeovsa
from suncasa.suncasatasks import importeovsa

import sys
import numpy as np
from eovsapy.dump_tsys import findfiles
from eovsapy.util import Time
import os
from suncasa.eovsa import eovsa_diskmodel as ed
from suncasa.utils import mstools as mstl

from suncasa.casa_compat import import_casatasks, import_casatools

tasks = import_casatasks('split', 'tclean', 'gencal', 'clearcal', 'applycal', 'gaincal',
                         'delmod')
split = tasks.get('split')
tclean = tasks.get('tclean')
gencal = tasks.get('gencal')
clearcal = tasks.get('clearcal')
applycal = tasks.get('applycal')
gaincal = tasks.get('gaincal')
delmod = tasks.get('delmod')

tools = import_casatools(['qatool', 'iatool', 'mstool', 'tbtool'])
qatool = tools['qatool']
iatool = tools['iatool']
mstool = tools['mstool']
tbtool = tools['tbtool']
ms = mstool()
tb = tbtool()

import socket
hostname = socket.gethostname()
import os

class Path_config:
    def __init__(self):
        self.udbmsdir = self._get_env_var('EOVSAUDBMS', '/data1/eovsa/fits/UDBms/')
        self.udbmsscldir = self._get_env_var('EOVSAUDBMSSCL', '/data1/eovsa/fits/UDBms_scl/')
        self.udbmsslfcaleddir = self._get_env_var('EOVSAUDBMSSLFCALED', '/data1/eovsa/fits/UDBms_slfcaled/')
        self.udbdir = self._get_env_var('EOVSAUDB', '/data1/eovsa/fits/UDB/')
        self.caltbdir = self._get_env_var('EOVSACAL', '/data1/eovsa/caltable/')
        self.slfcaltbdir = self._get_env_var('EOVSASLFCAL', '/data1/eovsa/slfcaltable/')
        self.qlookfitsdir = self._get_env_var('EOVSAQLOOKFITS', '/data1/eovsa/fits/synoptic/')
        self.qlookfigdir = self._get_env_var('EOVSAQLOOKFIG', '/common/webplots/qlookimg_10m/')
        self.synopticfigdir = self._get_env_var('EOVSASYNOPTICFIG', '/common/webplots/SynopticImg/')

    def _get_env_var(self, env_var, default_path):
        path = os.getenv(env_var) or default_path
        if not os.path.exists(path):
            if hostname!='pipeline':
                path = os.path.basename(default_path.rstrip('/')) + '/'
            os.makedirs(path)
        return path

# Usage
pathconfig = Path_config()
print(pathconfig.udbmsdir)  # Accessing the directory path

udbmsdir = pathconfig.udbmsdir
udbmsscldir = pathconfig.udbmsscldir
udbmsslfcaleddir = pathconfig.udbmsslfcaleddir
udbdir = pathconfig.udbdir
caltbdir = pathconfig.caltbdir
slfcaltbdir = pathconfig.slfcaltbdir
qlookfitsdir = pathconfig.qlookfitsdir
qlookfigdir = pathconfig.qlookfigdir
synopticfigdir = pathconfig.synopticfigdir

def getspwfreq(vis):
    '''

    :param vis:
    :return: mid frequencies in GHz of each spw in the vis
    '''
    tb.open(vis + '/SPECTRAL_WINDOW')
    reffreqs = tb.getcol('REF_FREQUENCY')
    bdwds = tb.getcol('TOTAL_BANDWIDTH')
    cfreqs = reffreqs + bdwds / 2.
    tb.close()
    cfreqs = cfreqs / 1.0e9
    return cfreqs


def trange2ms(trange=None, doimport=False, verbose=False, doscaling=False, overwrite=True):
    '''This finds all solar UDBms files within a timerange; If the UDBms file does not exist 
       in EOVSAUDBMSSCL, create one by calling importeovsa
       Required inputs:
       trange - can be 1) a single string or Time() object in UTC: use the entire day, e.g., '2017-08-01' or Time('2017-08-01')
                          if just a date, find all scans withing the same date in local time. 
                          if a complete time stamp, find the local date first (which may be different from that provided, 
                            and return all scans within that day
                       2) a range of Time(), e.g., Time(['2017-08-01 00:00','2017-08-01 23:00'])
                       3) None -- use current date Time.now()
       doimport - Boolean. If true, call importeovsa to import UDB files that are missing from 
                  those found in the directory specified in EOVSAUDBMSSCL. Otherwise, return
                  a list of ms files it has found.
       doscaling - Boolean. If true, scale cross-correlation amplitudes by using auto-correlations
       verbose - Boolean. If true, return more information
    '''
    import glob
    if trange is None:
        trange = Time.now()
    if type(trange) == list or type(trange) == str:
        try:
            trange = Time(trange)
        except:
            print('trange format not recognised. Abort....')
            return None

    # # in case of a single Time object was passed, adjusting the time range to start at the local beginning of the day.
    # # Initially, 'trange' is set to start at 20:00:00 UTC of the given day.
    # # After adjustment, 'trange' spans from 08:00:00 UTC of the same day (start of the local day)
    # # to 08:00:00 UTC of the following day, covering the entire local day.
    # # Example of change:
    # # Before: trange = "yyyy-mm-dd 20:00:00.000" (single starting point)
    # # After:  trange = ["yyyy-mm-dd 08:00:00.000", "yyyy-mm-dd+1 08:00:00.000"] (full day range)
    try:
        # if single Time object, the following line would report an error
        nt = len(trange)
        if len(trange) > 1:
            # more than one value
            trange = Time([trange[0], trange[-1]])
            tdatetime = trange[0].to_datetime()
        else:
            # single value in a list
            if trange[0].mjd == np.fix(trange[0].mjd):
                # if only date is given, move the time from 00 to 12 UT
                trange[0] = Time(trange[0].mjd + 0.5, format='mjd')

            tdatetime = trange[0].to_datetime()
            dhr = trange[0].LocalTime.utcoffset().total_seconds() / 60 / 60 / 24
            btime = Time(np.fix(trange[0].mjd + dhr) - dhr, format='mjd')
            etime = Time(btime.mjd + 1, format='mjd')
            trange = Time([btime, etime])
    except:
        # the case of a single Time object
        if trange.mjd == np.fix(trange.mjd):
            # if only date is given, move the time from 00 to 12 UT
            trange = Time(trange.mjd + 0.5, format='mjd')

        tdatetime = trange.to_datetime()
        dhr = trange.LocalTime.utcoffset().total_seconds() / 60 / 60 / 24
        btime = Time(np.fix(trange.mjd + dhr) - dhr, format='mjd')
        etime = Time(btime.mjd + 1, format='mjd')
        trange = Time([btime, etime])

    print('Selected timerange in UTC: ', trange.iso)

    if doscaling:
        udbmspath = udbmsscldir
    else:
        udbmspath = udbmsdir
    inpath = '{}{}/'.format(udbdir, tdatetime.strftime("%Y"))
    outpath = '{}{}/'.format(udbmspath, tdatetime.strftime("%Y%m"))
    if not os.path.exists(outpath):
        if verbose:
            print(outpath + ' does not exist. Making a new directory.')
        os.makedirs(outpath)
        msfiles = []
    else:
        msfiles = [os.path.basename(ll).split('.')[0] for ll in glob.glob('{}UDB*.ms'.format(outpath))]

    msfile_synoptic = os.path.join(outpath, 'UDB' + tdatetime.strftime("%Y%m%d") + '.ms')
    if overwrite and doimport:
        if os.path.exists(msfile_synoptic):
            os.system('rm -rf {}'.format(msfile_synoptic))

    # sclist = ra.findfiles(trange, projid='NormalObserving', srcid='Sun')
    sclist = findfiles(trange, projid='NormalObserving', srcid='Sun')
    udbfilelist = sclist['scanlist']
    udbfilelist = [os.path.basename(ll) for ll in udbfilelist]

    if os.path.exists(msfile_synoptic):
        return {'mspath': outpath, 'udbpath': inpath, 'udbfile': sorted(udbfilelist), 'udb2ms': [],
                'ms': [msfile_synoptic],
                'tstlist': sclist['tstlist'], 'tedlist': sclist['tedlist']}
    else:
        udbfilelist_set = set(udbfilelist)
        msfiles = udbfilelist_set.intersection(msfiles)
        filelist = udbfilelist_set - msfiles
        filelist = sorted(list(filelist))
        if filelist and doimport:
            # import multiprocessing as mprocs
            # ncpu = mprocs.cpu_count()
            # if ncpu > 10:
            #    ncpu = 10
            # if ncpu > len(filelist):
            #    ncpu = len(filelist)
            ncpu = 1
            importeovsa(idbfiles=[inpath + ll for ll in filelist], ncpu=ncpu, timebin="0s", width=1,
                        visprefix=outpath, nocreatms=False,
                        doconcat=False, modelms="", doscaling=doscaling, keep_nsclms=False, udb_corr=True)

        msfiles = [os.path.basename(ll).split('.')[0] for ll in glob.glob('{}UDB*.ms'.format(outpath))]
        udbfilelist_set = set(udbfilelist)
        msfiles = udbfilelist_set.intersection(msfiles)
        filelist = udbfilelist_set - msfiles
        filelist = sorted(list(filelist))

        return {'mspath': outpath, 'udbpath': inpath, 'udbfile': sorted(udbfilelist), 'udb2ms': filelist,
                'ms': [outpath + ll + '.ms' for ll in sorted(list(msfiles))], 'tstlist': sclist['tstlist'],
                'tedlist': sclist['tedlist']}


def calib_pipeline(trange, workdir=None, doimport=False, overwrite=False, clearcache=False, verbose=False, pols='XX',
                   version='v1.0', ncpu='auto'):
    ''' 
       trange: can be 1) a single Time() object: use the entire day
                      2) a range of Time(), e.g., Time(['2017-08-01 00:00','2017-08-01 23:00'])
                      3) a single or a list of UDBms file(s)
                      4) None -- use current date Time.now()
    '''

    if workdir is None:
        workdir = '/data1/workdir'
    os.chdir(workdir)
    if isinstance(trange, Time):
        mslist = trange2ms(trange=trange, doimport=False)
        invis = mslist['ms']
    if type(trange) == str:
        try:
            mslist = trange2ms(trange=trange, doimport=False)
            invis = mslist['ms']
        except:
            invis = [trange]

    for idx, f in enumerate(invis):
        if f[-1] == '/':
            invis[idx] = f[:-1]

    if overwrite or (invis == []):
        if isinstance(trange, Time):
            mslist = trange2ms(trange=trange, doimport=doimport, overwrite=overwrite)
            invis = mslist['ms']
        if type(trange) == str:
            try:
                mslist = trange2ms(trange=trange, doimport=doimport, overwrite=overwrite)
                invis = mslist['ms']
            except:
                invis = [trange]

        for idx, f in enumerate(invis):
            if f[-1] == '/':
                invis[idx] = f[:-1]

        outputvis = os.path.join(os.path.dirname(invis[0]), os.path.basename(invis[0])[:11] + '.ms')
        vis = calibeovsa(invis, caltype=['refpha', 'phacal'], caltbdir=caltbdir, interp='nearest',
                         doflag=True,
                         flagant='13~15',
                         doimage=False, doconcat=True,
                         concatvis=outputvis, keep_orig_ms=False)
    else:
        vis = invis[0]

    udbmspath = udbmsslfcaleddir
    tdate = mstl.get_trange(vis)[0]
    outpath = os.path.join(udbmspath, tdate.datetime.strftime('%Y%m')) + '/'
    if not os.path.exists(outpath):
        os.makedirs(outpath)
    imgoutdir = os.path.join(qlookfitsdir, tdate.datetime.strftime("%Y/%m/%d/"))
    if not os.path.exists(imgoutdir):
        os.makedirs(imgoutdir)
    figoutdir = os.path.join(synopticfigdir, tdate.datetime.strftime("%Y/"))
    if not os.path.exists(figoutdir):
        os.makedirs(figoutdir)

    if version == 'v1.0':
        output_file_path = outpath + os.path.basename(invis[0])[:11] + '.ms'
    else:
        output_file_path = outpath + os.path.basename(invis[0])[:11] + f'.{version}.ms.'
    slfcaltbdir_path = os.path.join(slfcaltbdir, tdate.datetime.strftime('%Y%m')) + '/'

    if verbose:
        print('input of pipeline_run:')
        print({'vis': vis,
               'outputvis': output_file_path,
               'workdir': workdir,
               'slfcaltbdir': slfcaltbdir_path,
               'imgoutdir': imgoutdir,
               'figoutdir': figoutdir,
               'overwrite':overwrite,
               'clearcache':clearcache,
               'pols':pols,'ncpu':ncpu})
    if version == 'v1.0':
        vis = ed.pipeline_run(vis, outputvis=output_file_path,
                              workdir=workdir,
                              slfcaltbdir=slfcaltbdir_path,
                              imgoutdir=imgoutdir, figoutdir=figoutdir, clearcache=clearcache, pols=pols)
    else:
        from suncasa.eovsa import eovsa_synoptic_imaging_pipeline as esip
        vis = esip.pipeline_run(vis, outputvis=output_file_path,
                                workdir=workdir,
                                slfcaltbdir=slfcaltbdir_path,
                                imgoutdir=imgoutdir, figoutdir=figoutdir, clearcache=clearcache, pols=pols, ncpu=ncpu,overwrite=overwrite)
    return vis


def mk_qlook_image(trange, doimport=False, docalib=False, ncpu=10, twidth=12, stokes=None, antenna='0~12',
                   lowcutoff_freq=3.7, imagedir=None, spws=['1~5', '6~10', '11~15', '16~25'], toTb=True, overwrite=True,
                   doslfcal=False, verbose=False):
    '''
       trange: can be 1) a single Time() object: use the entire day
                      2) a range of Time(), e.g., Time(['2017-08-01 00:00','2017-08-01 23:00'])
                      3) a single or a list of UDBms file(s)
                      4) None -- use current date Time.now()
    '''
    antenna0 = antenna
    if isinstance(trange, Time):
        mslist = trange2ms(trange=trange, doimport=doimport)
        vis = mslist['ms']
        tsts = [l.to_datetime() for l in mslist['tstlist']]
    if isinstance(trange, str):
        try:
            date = Time(trange)
            mslist = trange2ms(trange=trange, doimport=doimport)
            vis = mslist['ms']
            tsts = [l.to_datetime() for l in mslist['tstlist']]
        except:
            vis = [trange]
            tsts = []
            for v in vis:
                tb.open(v + '/OBSERVATION')
                tsts.append(Time(tb.getcell('TIME_RANGE')[0] / 24 / 3600, format='mjd').datetime)
                tb.close()
    subdir = [tst.strftime("%Y/%m/%d/") for tst in tsts]

    for idx, f in enumerate(vis):
        if f[-1] == '/':
            vis[idx] = f[:-1]
    if not stokes:
        stokes = 'XX'

    if not imagedir:
        imagedir = './'
    imres = {'Succeeded': [], 'BeginTime': [], 'EndTime': [], 'ImageName': [], 'Spw': [], 'Vis': [],
             'Synoptic': {'Succeeded': [], 'BeginTime': [], 'EndTime': [], 'ImageName': [], 'Spw': [], 'Vis': []}}
    for n, msfile in enumerate(vis):
        msfilebs = os.path.basename(msfile)
        imdir = imagedir + subdir[n]
        if not os.path.exists(imdir):
            os.makedirs(imdir)
        if doslfcal:
            slfcalms = './' + msfilebs + '.xx'
            split(msfile, outputvis=slfcalms, datacolumn='corrected', correlation='XX')
        cfreqs = getspwfreq(msfile)
        for spw in spws:
            antenna = antenna0
            if spw == '':
                continue
            spwran = [s.zfill(2) for s in spw.split('~')]
            freqran = [cfreqs[int(s)] for s in spw.split('~')]
            cfreq = np.mean(freqran)
            bmsz = max(150. / cfreq, 20.)
            uvrange = '<10klambda'
            if doslfcal:
                slfcal_img = './' + msfilebs + '.slf.spw' + spw.replace('~', '-') + '.slfimg'
                slfcal_tb = './' + msfilebs + '.slf.spw' + spw.replace('~', '-') + '.slftb'
                try:
                    tclean(vis=slfcalms, antenna=antenna, imagename=slfcal_img, spw=spw, mode='mfs', timerange='',
                           deconvolver='hogbom',
                           imsize=[512, 512], cell=['5arcsec'], niter=100, gain=0.05, stokes='I',
                           weighting='natural',
                           restoringbeam=[str(bmsz) + 'arcsec'], pbcor=False, interactive=False, usescratch=True)
                except:
                    print('error in cleaning spw: ' + spw)
                    break
                gaincal(vis=slfcalms, refant='0', antenna=antenna, caltable=slfcal_tb, spw=spw, uvrange='',
                        gaintable=[], selectdata=True,
                        timerange='', solint='600s', gaintype='G', calmode='p', combine='', minblperant=3, minsnr=2,
                        append=False)
                if not os.path.exists(slfcal_tb):
                    print('No solution found in spw: ' + spw)
                    break
                else:
                    clearcal(slfcalms)
                    delmod(slfcalms)
                    applycal(vis=slfcalms, gaintable=[slfcal_tb], spw=spw, selectdata=True, antenna=antenna,
                             interp='nearest', flagbackup=False,
                             applymode='calonly', calwt=False)
                    msfile = slfcalms

            imsize = 512
            cell = ['5arcsec']
            if len(spwran) == 2:
                spwstr = spwran[0] + '~' + spwran[1]
            else:
                spwstr = spwran[0]

            restoringbeam = ['{0:.1f}arcsec'.format(bmsz)]
            imagesuffix = '.spw' + spwstr.replace('~', '-')
            if cfreq > 10.:
                antenna = antenna + ';!0&1;!0&2'  # deselect the shortest baselines
            # else:
            #     antenna = antenna + ';!0&1'  # deselect the shortest baselines

            res = ptclean(vis=msfile, imageprefix=imdir, imagesuffix=imagesuffix, twidth=twidth, uvrange=uvrange,
                          spw=spw, ncpu=ncpu, niter=1000,
                          gain=0.05, antenna=antenna, imsize=imsize, cell=cell, stokes=stokes, doreg=True,
                          usephacenter=False, overwrite=overwrite,
                          toTb=toTb, restoringbeam=restoringbeam, specmode="mfs", deconvolver="hogbom",
                          datacolumn='data', pbcor=True)

            if res:
                imres['Succeeded'] += res['Succeeded']
                imres['BeginTime'] += res['BeginTime']
                imres['EndTime'] += res['EndTime']
                imres['ImageName'] += res['ImageName']
                imres['Spw'] += [spwstr] * len(res['ImageName'])
                imres['Vis'] += [msfile] * len(res['ImageName'])
            else:
                continue

    if len(vis) == 1:
        # produce the band-by-band whole-day images
        ms.open(msfile)
        ms.selectinit()
        timfreq = ms.getdata(['time', 'axis_info'], ifraxis=True)
        tim = timfreq['time']
        ms.done()

        cfreqs = getspwfreq(msfile)
        imdir = imagedir + subdir[0]
        if not os.path.exists(imdir):
            os.makedirs(imdir)
        for spw in spws:
            antenna = antenna0
            if spw == '':
                spw = '{:d}~{:d}'.format(next(x[0] for x in enumerate(cfreqs) if x[1] > lowcutoff_freq),
                                         len(cfreqs) - 1)
            spwran = [s.zfill(2) for s in spw.split('~')]
            freqran = [cfreqs[int(s)] for s in spw.split('~')]
            cfreq = np.mean(freqran)
            bmsz = max(150. / cfreq, 20.)
            uvrange = ''
            imsize = 512
            cell = ['5arcsec']
            if len(spwran) == 2:
                spwstr = spwran[0] + '~' + spwran[1]
            else:
                spwstr = spwran[0]

            restoringbeam = ['{0:.1f}arcsec'.format(bmsz)]
            imagesuffix = '.synoptic.spw' + spwstr.replace('~', '-')
            antenna = antenna + ';!0&1'  # deselect the shortest baselines

            res = ptclean(vis=msfile, imageprefix=imdir, imagesuffix=imagesuffix, twidth=len(tim), uvrange=uvrange,
                          spw=spw, ncpu=1, niter=0,
                          gain=0.05, antenna=antenna, imsize=imsize, cell=cell, stokes=stokes, doreg=True,
                          usephacenter=False, overwrite=overwrite,
                          toTb=toTb, restoringbeam=restoringbeam, specmode="mfs", deconvolver="hogbom",
                          datacolumn='data', pbcor=True)
            if res:
                imres['Synoptic']['Succeeded'] += res['Succeeded']
                imres['Synoptic']['BeginTime'] += res['BeginTime']
                imres['Synoptic']['EndTime'] += res['EndTime']
                imres['Synoptic']['ImageName'] += res['ImageName']
                imres['Synoptic']['Spw'] += [spwstr] * len(res['ImageName'])
                imres['Synoptic']['Vis'] += [msfile] * len(res['ImageName'])
            else:
                continue

    # save it for debugging purposes
    np.savez('imres.npz', imres=imres)

    return imres


def plt_qlook_image(imres, figdir=None, verbose=True, synoptic=False):
    from matplotlib import pyplot as plt
    from sunpy import map as smap
    from sunpy import sun
    from matplotlib import colors
    import astropy.units as u
    from suncasa.utils import plot_mapX as pmX
    # from matplotlib import gridspec as gridspec

    if not figdir:
        figdir = './'

    nspw = len(set(imres['Spw']))
    plttimes = list(set(imres['BeginTime']))
    ntime = len(plttimes)
    # sort the imres according to time
    images = np.array(imres['ImageName'])
    btimes = Time(imres['BeginTime'])
    etimes = Time(imres['EndTime'])
    spws = np.array(imres['Spw'])
    suc = np.array(imres['Succeeded'])
    inds = btimes.argsort()
    images_sort = images[inds].reshape(ntime, nspw)
    btimes_sort = btimes[inds].reshape(ntime, nspw)
    suc_sort = suc[inds].reshape(ntime, nspw)
    if verbose:
        print('{0:d} figures to plot'.format(ntime))
    plt.ioff()
    fig = plt.figure(figsize=(8, 8))

    plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
    axs = []
    ims = []
    pltst = 0
    for i in range(ntime):
        plt.ioff()
        plttime = btimes_sort[i, 0]
        tofd = plttime.mjd - np.fix(plttime.mjd)
        suci = suc_sort[i]
        if not synoptic:
            if tofd < 16. / 24. or sum(
                    suci) < nspw - 2:  # if time of the day is before 16 UT (and 24 UT), skip plotting (because the old antennas are not tracking)
                continue
            else:
                if pltst == 0:
                    i0 = i
                    pltst = 1
        else:
            if pltst == 0:
                i0 = i
                pltst = 1
        if i == i0:
            if synoptic:
                timetext = fig.text(0.01, 0.98, plttime.iso[:10], color='w', fontweight='bold', fontsize=12, ha='left')
            else:
                timetext = fig.text(0.01, 0.98, plttime.iso[:19], color='w', fontweight='bold', fontsize=12, ha='left')
        else:
            if synoptic:
                timetext.set_text(plttime.iso[:10])
            else:
                timetext.set_text(plttime.iso[:19])
        if verbose:
            print('Plotting image at: ', plttime.iso)
        for n in range(nspw):
            plt.ioff()
            if i == i0:
                if nspw == 1:
                    ax = fig.add_subplot(111)
                else:
                    ax = fig.add_subplot(nspw / 2, 2, n + 1)
                axs.append(ax)
            else:
                ax = axs[n]
            image = images_sort[i, n]
            if suci[n] or os.path.exists(image):
                try:
                    eomap = smap.Map(image)
                except:
                    continue
                data = eomap.data
                sz = data.shape
                if len(sz) == 4:
                    data = data.reshape((sz[2], sz[3]))
                data[np.isnan(data)] = 0.0
                # add a basin flux to the image to avoid negative values
                data = data + 0.8e5
                data[data < 0] = 0.0
                data = np.sqrt(data)
                eomap = smap.Map(data, eomap.meta)
                # resample the image for plotting
                dim = u.Quantity([256, 256], u.pixel)
                eomap = eomap.resample(dim)
            else:
                # make an empty map
                data = np.zeros((256, 256))
                header = {"DATE-OBS": plttime.isot, "EXPTIME": 0., "CDELT1": 10., "NAXIS1": 256, "CRVAL1": 0.,
                          "CRPIX1": 128.5, "CUNIT1": "arcsec",
                          "CTYPE1": "HPLN-TAN", "CDELT2": 10., "NAXIS2": 256, "CRVAL2": 0., "CRPIX2": 128.5,
                          "CUNIT2": "arcsec", "CTYPE2": "HPLT-TAN",
                          "HGLT_OBS": sun.heliographic_solar_center(plttime)[1].value, "HGLN_OBS": 0.,
                          "RSUN_OBS": sun.solar_semidiameter_angular_size(plttime).value,
                          "RSUN_REF": sun.constants.radius.value,
                          "DSUN_OBS": sun.sunearth_distance(plttime).to(u.meter).value, }
                eomap = smap.Map(data, header)
            if i == i0:
                eomap_ = pmX.Sunmap(eomap)
                # im = eomap_.imshow(axes=ax, cmap='jet', norm=colors.LogNorm(vmin=0.1, vmax=1e8))
                im = eomap_.imshow(axes=ax, cmap='jet', norm=colors.Normalize(vmin=150, vmax=700))
                ims.append(im)
                if not synoptic:
                    eomap_.draw_limb(axes=ax)
                eomap_.draw_grid(axes=ax)
                ax.set_xlim([-1080, 1080])
                ax.set_ylim([-1080, 1080])
                try:
                    cfreq = eomap.meta['crval3'] / 1.0e9
                    bdwid = eomap.meta['cdelt3'] / 1.0e9
                    ax.text(0.98, 0.01, '{0:.1f} - {1:.1f} GHz'.format(cfreq - bdwid / 2.0, cfreq + bdwid / 2.0),
                            color='w', transform=ax.transAxes, fontweight='bold', ha='right')
                except:
                    pass
                ax.set_title(' ')
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.set_xticklabels([''])
                ax.set_yticklabels([''])
            else:
                ims[n].set_data(eomap.data)

        fig_tdt = plttime.to_datetime()
        if synoptic:
            fig_subdir = fig_tdt.strftime("%Y/")
            figname = 'eovsa_qlimg_' + plttime.iso[:10].replace('-', '') + '.png'
        else:
            fig_subdir = fig_tdt.strftime("%Y/%m/%d/")
            figname = 'eovsa_qlimg_' + plttime.isot.replace(':', '').replace('-', '')[:15] + '.png'
        figdir_ = figdir + fig_subdir
        if not os.path.exists(figdir_):
            os.makedirs(figdir_)
        if verbose:
            print('Saving plot to :' + figdir_ + figname)

        plt.savefig(figdir_ + figname)
    plt.close(fig)


def qlook_image_pipeline(date, twidth=10, ncpu=15, doimport=False, docalib=False, synoptic=False, overwrite=True):
    ''' date: date string or Time object. e.g., '2017-07-15' or Time('2017-07-15')
    '''
    import pytz
    from datetime import datetime
    if date is None:
        date = Time.now()
    try:
        date = Time(date)
    except:
        print('date format not recognised. Abort....')
        return None

    if date.mjd > Time('2019-02-02 12:00:00').mjd:
        ## the last '' window is for fullBD synthesis image. Now obsolete.
        # spws = ['6~10', '11~20', '21~30', '31~43', '']
        # spws = ['6~10', '11~20', '21~30', '31~43']
        spws = ['0~1', '2~5', '6~10', '11~20', '21~30', '31~49']
    else:
        ## the last '' window is for fullBD synthesis image. Now obsolete.
        # spws = ['1~5', '6~10', '11~15', '16~25', '']
        spws = ['1~3', '4~6', '7~10', '10~14', '15~20', '21~30']

    if docalib:
        vis = calib_pipeline(date, doimport=doimport, synoptic=synoptic)

    imagedir = qlookfitsdir
    if synoptic:
        vis_synoptic = os.path.join(udbmsdir, date.datetime.strftime("%Y%m"),
                                    'UDB' + date.datetime.strftime("%Y%m%d") + '.ms')
        if os.path.exists(vis_synoptic):
            date = vis_synoptic
        else:
            print('Whole-day ms file {} not existed. About..... Use pipeline to make one.'.format(vis_synoptic))
            return None

    imres = mk_qlook_image(date, twidth=twidth, ncpu=ncpu, doimport=doimport, docalib=docalib, imagedir=imagedir,
                           spws=spws, verbose=True, overwrite=overwrite)

    figdir = qlookfigdir
    plt_qlook_image(imres, figdir=figdir, verbose=True)
    # if imres['Synoptic']['Succeeded']:
    #     figdir = synopticfigdir
    #     imres_bds = {}
    #     # imres_allbd = {}
    #     for k, v in imres['Synoptic'].items():
    #         imres_bds[k] = v  # [:4]
    #         # imres_allbd[k] = v[4:]
    #
    #     plt_qlook_image(imres_bds, figdir=figdir, verbose=True, synoptic=True)
    #     # plt_qlook_image(imres_allbd, figdir=figdir + 'FullBD/', verbose=True, synoptic=True)


def pipeline(year=None, month=None, day=None, ndays=1, clearcache=True, overwrite=False, doimport=True, pols='XX',
             version='v1.0', ncpu='auto', debugging=False):
    """
    Main pipeline for importing and calibrating EOVSA visibility data.

    Name:
        eovsa_pipeline --- main pipeline for importing and calibrating EOVSA visibility data.

    Synopsis:
        eovsa_pipeline.py [options]... [DATE_IN_YY_MM_DD]

    Description:
        Import and calibrate EOVSA visibility data of the date specified
        by DATE_IN_YY_MM_DD (or from ndays before the DATE_IN_YY_MM_DD if option --ndays/-n is provided).
        If DATE_IN_YY_MM_DD is omitted, it will be set to 2 days before now by default.
        There are no mandatory arguments in this command.

    :param year: The year for which data should be processed, defaults to None.
    :type year: int, optional
    :param month: The month for which data should be processed, defaults to None.
    :type month: int, optional
    :param day: The day for which data should be processed, defaults to None.
    :type day: int, optional
    :param ndays: Number of days before the specified date to include in the processing, defaults to 1.
    :type ndays: int, optional
    :param clearcache: Whether to clear cache after processing, defaults to True.
    :type clearcache: bool, optional
    :param overwrite: Whether to overwrite existing files, defaults to True.
    :type overwrite: bool, optional
    :param doimport: Whether to perform the import step, defaults to True.
    :type doimport: bool, optional
    :param pols: Polarizations to process, can be 'XX' or 'XXYY', defaults to 'XX'.
    :type pols: str, optional
    :param version: Version of the pipeline to use, choices are 'v1.0' or 'v2.0', defaults to 'v1.0'.
    :type version: str, optional
    :param ncpu: Number of CPUs to use for processing, defaults to 'auto'.
    :type ncpu: str, optional
    :param debugging: Whether to run the pipeline in debugging mode, defaults to False.
    :type debugging: bool, optional

    :raises ValueError: Raises an exception if the date parameters are out of the valid Gregorian calendar range.

    Example:
    --------
    To process data for November 24th, 2021 using version 2.0 of the pipeline, with all options enabled:

    >>> python eovsa_pipeline.py --date 2021-11-24T20:00 --clearcache --overwrite --doimport --pols XX --version v2.0 --ndays 2

    If you want to see the help message, you can run:

    >>> python eovsa_pipeline.py -h
    """
    workdir = '/data1/workdir/'
    os.chdir(workdir)
    if year is None:
        # Default behavior: Process data from one day prior to the current date.
        # Calculate the Modified Julian Date (MJD) for yesterday.
        mjdnow = Time.now().mjd - 1
        # Convert MJD to a datetime object and format it to start processing at 20:00 UT.
        t = Time(Time(mjdnow, format='mjd').to_datetime().strftime('%Y-%m-%dT20:00'))
    else:
        t = Time('{}-{:02d}-{:02d} 20:00'.format(year, month, day))
    for d in range(ndays):
        t1 = Time(t.mjd - d, format='mjd')
        datestr = t1.iso[:10]
        subdir = os.path.join(workdir, t1.datetime.strftime('%Y%m%d/'))
        if not os.path.exists(subdir):
            os.makedirs(subdir)
        else:
            if overwrite:
                os.system('rm -rf {}/*'.format(subdir))
        # ##debug
        # vis_corrected = calib_pipeline(datestr, overwrite=overwrite, doimport=doimport,
        #                                workdir=subdir, clearcache=False, pols=pols)

        if debugging:
            vis_corrected = calib_pipeline(t1, overwrite=overwrite, doimport=doimport,
                                           workdir=subdir, clearcache=False, pols=pols, version=version, ncpu=ncpu)
        else:
            try:
                vis_corrected = calib_pipeline(t1, overwrite=overwrite, doimport=doimport,
                                               workdir=subdir, clearcache=False, pols=pols, version=version, ncpu=ncpu)
            except Exception as e:
                print(f'error in processing {datestr}. Error message: {e}')
        if clearcache:
            os.chdir(workdir)
            os.system('rm -rf {}'.format(subdir))


if __name__ == '__main__':
    # Define the parser
    parser = argparse.ArgumentParser(description='EOVSA Pipeline for importing and calibrating visibility data.')
    # Default date is set to one day before the current date, formatted as YYYY-MM-DDT20:00
    default_date = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%dT20:00')
    parser.add_argument('--date', type=str, default=default_date,
                        help='Date to process in YYYY-MM-DDT20:00 format, defaults to 20:00 UT of one day before the current date.')
    parser.add_argument('--clearcache', action='store_true', default=False,
                        help='Remove temporary files after processing')
    parser.add_argument('--ndays', type=int, default=1,
                        help='Process data from DATE_IN_YY_MM_DD-ndays to DATE_IN_YY_MM_DD, default is 1.')
    parser.add_argument('--overwrite', action='store_true', default=False, help='Overwrite existing processed data')
    parser.add_argument('--doimport', action='store_true', default=True, help='Perform import step before processing')
    parser.add_argument('--pols', type=str, default='XX', choices=['XX', 'XXYY'], help='Polarizations to process')
    parser.add_argument('--ncpu', type=str, default='auto', help='Number of CPUs to use for processing')
    parser.add_argument('--version', type=str, default='v1.0', choices=['v1.0', 'v2.0'],
                        help='Version of the EOVSA pipeline to use')
    parser.add_argument('--debugging', action='store_true', default=False, help='Run the pipeline in debugging mode')

    # Parse the arguments
    args = parser.parse_args()

    # Convert --date argument to an astropy Time object to ensure consistency in time handling
    t = Time(args.date)

    # Extract year, month, day from the --date argument
    year, month, day = t.datetime.year, t.datetime.month, t.datetime.day

    # Run the main pipeline function
    pipeline(year, month, day, args.ndays, args.clearcache, args.overwrite, args.doimport, args.pols,
             args.version, args.ncpu, args.debugging)



__all__ = ['Dspec']

import os
import struct

import matplotlib.pyplot as plt
import numpy as np
from astropy.time import Time
from copy import copy

from ..casa_compat import import_casatools, import_casatasks
tasks = import_casatasks('split','hanningsmooth')
split = tasks.get('split')
hanningsmooth = tasks.get('hanningsmooth')

tools = import_casatools(['tbtool', 'mstool', 'qatool'])

tbtool = tools['tbtool']
mstool = tools['mstool']
qatool = tools['qatool']
tb = tbtool()
ms = mstool()
qa = qatool()

from matplotlib.dates import AutoDateFormatter, AutoDateLocator, num2date
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .sources.lwa import rebin1d, rebin2d

stokestype = [
    'Undefined',
    # undefined value = 0
    'I',
    'Q',
    'U',
    # standard stokes parameters
    'V',
    #
    'RR',
    'RL',
    'LR',
    # circular correlation products
    'LL',
    #
    'XX',
    'XY',
    'YX',
    # linear correlation products
    'YY',
    #
    'RX',
    'RY',
    'LX',
    'LY',
    'XR',
    'XL',
    'YR',
    # mixed correlation products
    'YL',
    #
    'PP',
    'PQ',
    'QP',
    # general quasi-orthogonal correlation products
    'QQ',
    #
    'RCircular',
    'LCircular',
    # single dish polarization types
    'Linear',
    # Polarized intensity ((Q^2+U^2+V^2)^(1/2))
    'Ptotal',
    # Linearly Polarized intensity ((Q^2+U^2)^(1/2))
    'Plinear',
    # Polarization Fraction (Ptotal/I)
    'PFtotal',
    # Linear Polarization Fraction (Plinear/I)
    'PFlinear',
    # Linear Polarization Angle (0.5 arctan(U/Q)) (in radians)
    'Pangle']

stokesenum = {}
for k, v in zip(range(len(stokestype)), stokestype):
    stokesenum[k] = v


def calc_bkg_dspec(spec, tim, bkgtim, interp_method='linear'):
    """
    Calculate the background spectrum.

    This function computes the background spectrum from a given spectrogram
    over specified time intervals. The background spectrum can be interpolated
    using various methods.

    :param spec: A 2D array of shape (nfreq, ntime) representing the spectrogram.
    :type spec: ndarray
    :param tim: A time object with the time axis of shape (ntime,).
    :type tim: astropy.time.Time
    :param bkgtim: Background time interval(s). It can be a single time interval (list of two strings)
        or multiple time intervals (list of lists of two strings each).
        Example: ['2024-06-11T19:41:00', '2024-06-11T20:40:58']
                 [['2024-06-11T19:41:00', '2024-06-11T20:40:58'], ['2024-06-11T21:41:00', '2024-06-11T22:40:58']]
    :type bkgtim: list or list of lists
    :param interp_method: The method of interpolation to use. Default is 'linear'.
        Can be any method supported by scipy.interpolate.interp1d.
    :type interp_method: str, optional
    :return: A 2D array of shape (nfreq, ntime) representing the interpolated background spectrum.
    :rtype: ndarray

    :example:
    >>> from astropy.time import Time
    >>> import numpy as np
    >>> spec = np.random.random((100, 200))
    >>> tim = Time(['2024-06-11T19:00:00']*200, scale='utc')
    >>> bkgtim = ['2024-06-11T19:41:00', '2024-06-11T20:40:58']
    >>> calc_bkg_dspec(spec, tim, bkgtim)
    array([[0.52203423, 0.52203423, 0.52203423, ..., 0.52203423, 0.52203423, 0.52203423],
           [0.48498126, 0.48498126, 0.48498126, ..., 0.48498126, 0.48498126, 0.48498126],
           ...,
           [0.49162387, 0.49162387, 0.49162387, ..., 0.49162387, 0.49162387, 0.49162387]])
    """
    from scipy.interpolate import interp1d
    from astropy.time import Time

    # Ensure tim is an array of julian dates
    tim_jd = tim.jd
    bkgtim_jd = Time(bkgtim).jd

    if isinstance(bkgtim[0], str):
        # Single time interval
        tidx_bkg, = np.where((tim_jd >= bkgtim_jd[0]) & (tim_jd <= bkgtim_jd[1]))
        spec_bkg = np.nanmean(spec[:, tidx_bkg], axis=1)[:, np.newaxis]
    else:
        # Multiple time intervals
        all_spec_bkg = []
        for bkgtjd in bkgtim_jd:
            tidx_bkg, = np.where((tim_jd >= bkgtjd[0]) & (tim_jd <= bkgtjd[1]))
            all_spec_bkg.append(np.nanmean(spec[:, tidx_bkg], axis=1))

        # Combine all background spectra
        all_spec_bkg = np.array(all_spec_bkg)  # Shape will be (nfreq, n_intervals)
        bkg_times = np.nanmean(bkgtim_jd,axis=1)

        interp_func = interp1d(bkg_times, all_spec_bkg, kind=interp_method, axis=0, fill_value="extrapolate")
        spec_bkg = interp_func(tim_jd).T  # Shape will be (nfreq, ntime)

    return spec_bkg

class Dspec:
    """
    A class to handle dynamic spectra from radio observations.

    Attributes
    ----------
    data : numpy.ndarray
        The dynamic spectrum data array.
    time_axis : astropy.time.Time
        Time axis of the dynamic spectrum.
    freq_axis : numpy.ndarray
        Frequency axis of the dynamic spectrum in Hz.
    telescope : str
        Name of the telescope used for the observation.
    observatory : str
        Name of the observatory.
    t_label : str
        Label for the time axis.
    f_label : str
        Label for the frequency axis.
    bl : str
        Baseline information.
    uvrange : str
        UV range information.
    pol : str
        Polarization information.
    spec_unit : str, default 'sfu'
        Unit of the dynamic spectrum data ('sfu', 'Jy', or 'K').

    Methods
    -------
    __init__(fname=None, specfile=None, **kwargs):
        Initializes the Dspec object by reading a FITS file or a saved numpy array.

    read(fname, source=None, *args, **kwargs):
        Reads dynamic spectrum data from a file.

    tofits(fitsfile=None, specdata=None, **kwargs):
        Writes the dynamic spectrum data to a FITS file.

    get_dspec(fname=None, **kwargs):
        Extracts dynamic spectrum data from a measurement set.

    wrt_dspec(specfile=None, specdat=None):
        Writes the dynamic spectrum data to a binary file.

    rd_dspec(specdata, **kwargs):
        Reads dynamic spectrum data from a numpy file.

    concat_dspec(specfiles, outfile=None, savespec=False):
        Concatenates multiple dynamic spectrum files along the time axis.

    peek(*args, **kwargs):
        Plots the dynamic spectrum on the current axes.

    plot(pol='I', **kwargs):
        Plots the dynamic spectrum for a given polarization.

    """
    data = None
    time_axis = None
    freq_axis = None
    telescope = None
    observatory = None
    t_label = None
    f_label = None
    bl = None
    uvrange = None
    pol = None
    spec_unit = 'sfu'

    def __init__(self, fname=None, specfile=None, bl='', uvrange='', field='', scan='',
                 datacolumn='data', domedian=False, timeran=None, spw=None, timebin='0s', regridfreq=False,
                 fillnan=None, verbose=False, usetbtool=True, ds_normalised=False):
        """
              Initializes the Dspec object by reading a FITS file or a saved numpy array.

              Parameters
              ----------
              fname : str, optional
                  The file name of the FITS file or the measurement set to be read.
              specfile : str, optional
                  The file name of the saved numpy array (.npz file) that contains the dynamic spectrum data.
              bl : str, optional
                  Baseline selection criteria.
              uvrange : str, optional
                  UV range selection criteria.
              field : str, optional
                  Field selection criteria.
              scan : str, optional
                  Scan selection criteria.
              datacolumn : str, optional
                  Specifies which data column ('data' or 'corrected') to read from the measurement set.
              domedian : bool, optional
                  If True, performs a median over the baseline axis.
              timeran : str or list of str, optional
                  Time range to select data.
              spw : str, optional
                  Spectral window selection criteria.
              timebin : str, optional
                  Time averaging interval.
              regridfreq : bool, optional
                  If True, regrids the frequency axis to have a uniform channel width.
              fillnan : float or int, optional
                  Value to fill in for flagged or NaN values.
              verbose : bool, optional
                  If True, prints detailed information during operation.
              usetbtool : bool, optional
                  If True, uses CASA table tools for data extraction.
              ds_normalised : bool, optional
                  If True, normalizes the dynamic spectrum by the median value.

              """
        if fname:
            if isinstance(fname, str):
                if os.path.exists(fname):
                    if os.path.isdir(fname):
                        self.get_dspec(fname, specfile=specfile, bl=bl, uvrange=uvrange, field=field,
                                       scan=scan, datacolumn=datacolumn, domedian=domedian, timeran=timeran, spw=spw,
                                       timebin=timebin, regridfreq=regridfreq, fillnan=fillnan, verbose=verbose,
                                       usetbtool=usetbtool, ds_normalised=ds_normalised)
                    else:
                        self.read(fname)
                else:
                    raise FileNotFoundError(f"The file {fname} does not exist. Please check your file path.")
            else:
                self.read(fname)

    def read(self, fname, source=None, *args, **kwargs):
        """
        Reads dynamic spectrum data from a file.

        Parameters
        ----------
        fname : str or list of str
            The file name to read the dynamic spectrum data from.
        source : str, optional
            Specifies the data source ('fits', 'suncasa', or 'lwa') to determine the appropriate reader.

        Additional parameters are passed to the specific reader function based on the source.

        """
        ##TODO The existing implementation of the mapping between extensions and instruments requires refinement and sophistication.
        # Explore potential optimization strategies to improve this process.
        from astropy.io import fits

        if type(fname) is str:
            _known_extensions = {
                ('npz'): 'suncasa',
            }
            for extension, readername in _known_extensions.items():
                if fname.lower().endswith(extension):
                    source = readername
            if source is None and (not fname.lower().startswith('eovsa')):
                raise ValueError(f"The filetype provided ({os.path.basename(fname)}) is not supported")

            if source.lower() == 'eovsa' or fname.lower().startswith('eovsa'):
                if fname.endswith('.fts') or fname.endswith('.fits'):
                    from .sources import eovsa
                    s = eovsa.get_dspec(fname, doplot=False)
                    self.data = s['spectrogram']
                    self.time_axis = Time(s['time_axis'], format='mjd')
                    self.freq_axis = s['spectrum_axis'] * 1e9
                    self.spec_unit = 'sfu'
                    self.telescope = 'EOVSA'
                    self.observatory = 'OVRO'

            if source.lower() == 'suncasa':
                spec, tim, freq, bl, pol, spec_unit = self.rd_dspec(fname, spectype='amp', spec_unit='jy')
                self.data = spec
                self.time_axis = Time(tim / 24 / 3600, format='mjd')
                self.freq_axis = freq
                self.bl = bl
                self.pol = pol
                self.spec_unit = spec_unit
                self.telescope = ''
                self.observatory = ''

            if source.lower() == 'lwa' and fname.endswith('.fits'):
                    hdu = fits.open(fname) 
                    self.data = hdu[0].data
                    tim = hdu[2].data
                    tmjd = np.array(tim['mjd']) + np.array(tim['time']) / 24. / 3600 / 1000
                    self.time_axis = Time(tmjd, format='mjd')
                    self.freq_axis = hdu[1].data['sfreq'] * 1e9
                    self.pol = [hdu[0].header['POLARIZA']]
                    self.spec_unit = 'sfu'
                    self.telescope = 'LWA'
                    self.observatory = 'OVRO'

        elif source.lower() == 'lwa' and type(fname) is list:
            from .sources import lwa
            spec, tim, freq, pol, calfac_x, calfac_y, bkg_flux = lwa.read_data(fname, **kwargs)
            self.data = spec
            self.time_axis = Time(tim, format='mjd')
            self.freq_axis = freq
            self.pol = pol
            self.calfac_x = calfac_x # correction factor for X pol, same shape as freq
            self.calfac_y = calfac_y # correction factor for Y pol, same shape as freq
            self.bkg_flux = bkg_flux
            self.spec_unit = 'sfu'
            self.telescope = 'LWA'
            self.observatory = 'OVRO'

        else:
            raise ValueError(f"Unsupported data source or type provided: {source}")

    def tofits(self, fitsfile=None, specdata=None, spectype='amp', spec_unit='jy',
               telescope='EOVSA', observatory='Owens Valley Radio Observatory', observer='EOVSA Team'):
        """
        Writes the dynamic spectrum data to a FITS file.

        This method exports the stored dynamic spectrum data into a new FITS file, including relevant metadata and optional input spectrum data.

        Parameters
        ----------
        fitsfile : str, optional
            Path and name of the output FITS file. If not specified, a default name will be generated.
        specdata : dict, optional
            Input dictionary containing dynamic spectrum data and metadata. If not provided, the method uses the data stored in the Dspec object.
        spectype : str, optional
            Specifies the type of the spectrum to be saved ('amp' for amplitude, 'pha' for phase). Default is 'amp'.
        spec_unit : str, optional
            Specifies the unit of the spectrum data ('jy' for Jansky, 'sfu' for Solar Flux Units, 'k' for Kelvin). Default is 'jy'.
        telescope : str, optional
            Name of the telescope with which the data was obtained. Default is 'EOVSA'.
        observatory : str, optional
            Name of the observatory. Default is 'Owens Valley Radio Observatory'.
        observer : str, optional
            Name of the observer or team that made the observation. Default is 'EOVSA Team'.

        Returns
        -------
        None

        Raises
        ------
        FileNotFoundError
            If the specified `fitsfile` path does not exist.
        ValueError
            If `specdata` is provided but does not contain the required keys.

        Examples
        --------
        >>> dspec = Dspec()
        >>> dspec.tofits(fitsfile='output.fits', specdata=my_specdata, spec_unit='sfu', observer='Sun Observer')

        Note
        ----
        The method can directly utilize the dynamic spectrum data stored within the `Dspec` object if `specdata` is not provided. Ensure the `Dspec` object has been properly initialized with dynamic spectrum data before calling this method without `specdata`.
        """
        from astropy.io import fits

        if hasattr(self, 'data'):
            spec = self.data
            tim = self.time_axis
            freqghz = self.freq_axis / 1e9

        if specdata:
            try:
                spec, tim, freq, bl, pol, spec_unit = self.rd_dspec(specdata, spectype=spectype, spec_unit=spec_unit)
                tim = Time(tim / 24 / 3600, format='mjd')
            except:
                print('Something is wrong with the input specdata.')

        hdu = fits.PrimaryHDU(spec)
        # Set up the extensions: sfreq, ut
        col1 = fits.Column(name='sfreq', format='E', array=freqghz)
        cols1 = fits.ColDefs([col1])
        tbhdu1 = fits.BinTableHDU.from_columns(cols1)
        tbhdu1.name = 'SFREQ'

        # Split up mjd into days and msec
        date_obs = tim[0].isot
        date_end = tim[-1].isot
        ut = tim.mjd
        ut_int = ut.astype(int)
        ut_msec = 1000.0 * 86400.0 * (ut - ut_int)
        ut_ms1 = ut_msec.astype(int)

        # J is the format code for a 32 bit integer, who would have thought
        # http://astropy.readthedocs.org/en/latest/io/fits/usage/table.html
        col3 = fits.Column(name='mjd', format='J', array=ut_int)
        col4 = fits.Column(name='time', format='J', array=ut_ms1)

        cols3 = fits.ColDefs([col3, col4])
        tbhdu3 = fits.BinTableHDU.from_columns(cols3)
        tbhdu3.name = 'UT'

        if hasattr(self, 'calfac_x') and hasattr(self, 'calfac_y') and hasattr(self, 'bkg_flux'):
            # add calfac_x _y to FREQ_CAL
            col5 = fits.Column(name='calfac_x', format='E', array=self.calfac_x)
            col6 = fits.Column(name='calfac_y', format='E', array=self.calfac_y)
            col7 = fits.Column(name='bkg_flux', format='E', array=self.bkg_flux)
            cols5 = fits.ColDefs([col5, col6, col7])
            tbhdu5 = fits.BinTableHDU.from_columns(cols5)
            tbhdu5.name = 'FREQ_CAL'
        else:
            tbhdu5 = fits.BinTableHDU.from_columns([])
            tbhdu5.name = 'FREQ_CAL'
            
        # create an HDUList object to put in header information
        hdulist = fits.HDUList([hdu, tbhdu1, tbhdu3, tbhdu5])

        # primary header
        prihdr = hdulist[0].header
        prihdr.set('FILENAME', fitsfile)
        prihdr.set('ORIGIN', 'NJIT', 'Location where file was made')
        prihdr.set('DATE', Time.now().isot, 'Date when file was made')
        prihdr.set('OBSERVER', observer, 'Who to appreciate/blame')
        prihdr.set('TELESCOP', telescope, observatory)
        prihdr.set('OBJ_ID', 'SUN', 'Object ID')
        prihdr.set('TYPE', 1, 'Flare Spectrum')
        prihdr.set('DATE_OBS', date_obs, 'Start date/time of observation')
        prihdr.set('DATE_END', date_end, 'End date/time of observation')
        prihdr.set('FREQMIN', min(freqghz), 'Min freq in observation (GHz)')
        prihdr.set('FREQMAX', max(freqghz), 'Max freq in observation (GHz)')
        prihdr.set('XCEN', 0.0, 'Antenna pointing in arcsec from Sun center')
        prihdr.set('YCEN', 0.0, 'Antenna pointing in arcsec from Sun center')
        prihdr.set('POLARIZA', 'I', 'Polarizations present')
        prihdr.set('RESOLUTI', 0.0, 'Resolution value')
        # Write the file
        hdulist.writeto(fitsfile, overwrite=True)

    def get_dspec(self, fname=None, specfile=None, bl='', uvrange='', field='', scan='',
                  datacolumn='data',
                  domedian=False, timeran=None, spw=None, timebin='0s', regridfreq=False,
                  hanning=False,
                  applyflag=True, fillnan=None, verbose=False,
                  usetbtool=True, ds_normalised=False):
        if fname.endswith('/'):
            fname = fname[:-1]
        msfile = fname
        if not spw:
            spw = ''
        if not timeran:
            timeran = ''
        if domedian:
            if not uvrange:
                uvrange = '0.2~0.8km'
        if not bl:
            bl = ''
        else:
            uvrange = ''
        # Open the ms and plot dynamic spectrum
        if verbose:
            print('Splitting selected data...')

        if usetbtool:
            if datacolumn.lower() == 'data':
                datacol = 'DATA'
                print('Using DATA column')
            if datacolumn.lower() == 'corrected':
                datacol = 'CORRECTED_DATA'
                print('Using CORRECTED_DATA column')
            if verbose:
                print('using table tool to extract the data')
            try:
                tb.open(fname + '/POLARIZATION')
                corrtype = tb.getcell('CORR_TYPE', 0)
                pol = [stokesenum[p] for p in corrtype]
                tb.close()
            except:
                pol = []

            if hanning:
                hanningsmooth(vis=fname, datacolumn='data', field=field, outputvis=fname + '.tmpms')
                fname = fname + '.tmpms'
            tb.open(fname)
            spwtb = tbtool()
            spwtb.open(fname + '/SPECTRAL_WINDOW')
            ptb = tbtool()
            ptb.open(fname + '/POLARIZATION')
            npol = ptb.getcol('NUM_CORR', 0, 1)[0]
            ptb.close()

            ms.open(fname)
            # ms.selectinit(datadescid=0)
            spwlist = []
            mdata = ms.metadata()
            # determine how many unique spws
            # now it ignores the spw parameter and use all of them
            spws_flds = mdata.spwsforfields()
            spws_ = []
            for k in spws_flds.keys():
                spws_.append(spws_flds[k])
            spws_unq = np.unique(spws_)
            nspw = len(spws_unq)
            nbaselines = mdata.nbaselines()
            nantennas = mdata.nantennas()
            scannumbers = mdata.scannumbers()
            spw_nfrq = []  # List of number of frequencies in each spw
            for i in spws_unq:
                spw_nfrq.append(mdata.nchan(i))
            spw_nfrq = np.array(spw_nfrq)
            nf = np.sum(spw_nfrq)
            smry = mdata.summary()

            scanids = sorted(smry['observationID=0']['arrayID=0'].keys())
            ## todo find a way to determine the target fieldID if multi-fields exist
            smryscan0 = smry['observationID=0']['arrayID=0'][scanids[0]]
            fieldid = ''
            for k in smryscan0.keys():
                if k.startswith('fieldID='):
                    fieldid = k
                    break
            nbl = int(smryscan0[fieldid]['0']['nrows'] / nspw)
            if nbl == nbaselines + nantennas:
                hasautocorr = True
            elif nbl == nbaselines:
                hasautocorr = False
            else:
                raise (ValueError('The baseline number is not correct.'))

            antmask = []
            if uvrange != '' or bl != '':
                ms.open(fname)
                ms.selectinit(datadescid=0, reset=True)
                mdata = ms.metadata()
                antlist = mdata.antennaids()
                smry = mdata.summary()
                mdata.done()
                staql = {'uvdist': uvrange, 'baseline': bl, 'spw': spw, 'field': field, 'scan': scan,
                         'timerange': timeran}
                ### todo the selection only works for uvrange and bl. To make the selection of other items works,
                a = ms.msselect(staql)
                mdata = ms.metadata()
                baselines = mdata.baselines()
                if hasautocorr:
                    for lidx, l in enumerate(antlist):
                        antmask.append(baselines[l][antlist[lidx:]])
                else:
                    for lidx, l in enumerate(antlist):
                        antmask.append(baselines[l][antlist[lidx + 1:]])
                antmask = np.hstack(antmask)
                mdata.done()
                ms.close()
                ms.open(fname)

            scan_ntimes = []  # List of number of times in each scan
            nrows = []
            for s, scanid in enumerate(scanids):
                smryscan = smry['observationID=0']['arrayID=0'][scanid]
                nrow = next(v for k, v in smryscan.items() if 'fieldID=' in k)['nrows']
                nrows.append(nrow)
                scan_ntimes.append(nrow / nspw / nbl)
            scan_ntimes = np.array(scan_ntimes)
            scan_ntimes_integer = scan_ntimes.astype(int)
            if len(np.where(scan_ntimes % scan_ntimes_integer != 0)[0]) != 0:
                # if True:
                scan_ntimes = []  # List of number of times in each scan
                for s, scanid in enumerate(scanids):
                    nrows_scan = []  ## get the nrows for each time. They are not always the SAME!
                    smryscan = smry['observationID=0']['arrayID=0'][scanid]
                    # for k, v in smryscan['fieldID=0'].items():
                    #     if isinstance(v,dict):
                    #         nrows_scan.append(v['nrows'])
                    nrows_scan.append(next(v for k, v in smryscan.items() if 'fieldID=' in k)['nrows'])
                    scan_ntimes.append(nrows[s] / max(set(nrows_scan)))
                    # scan_ntimes.append(
                    #     len(smry['observationID=0']['arrayID=0'][scanids[scannumber]]['fieldID=0'].keys()) - 6)
                scan_ntimes = np.array(scan_ntimes).astype(int)
            else:
                scan_ntimes = scan_ntimes_integer

            nt = np.sum(scan_ntimes)
            times = tb.getcol('TIME')
            if times[nbl] - times[0] != 0:
                # This is frequency/scan sort order
                order = 'f'
            elif times[nbl * nspw - 1] - times[0] != 0:
                # This is time sort order
                order = 't'
            else:
                order = 'f'

            freq = np.zeros(nf, float)
            times = np.zeros(nt, float)
            if verbose:
                print("npol, nf, nt, nbl:", npol, nf, nt, nbl)
            if order == 't':
                specamp = np.zeros((npol, nf, nbl, nt), complex)
                flagf = np.zeros((npol, nf, nbl, nt), int)
                for j in range(nt):
                    fptr = 0
                    # Loop over spw
                    for i, sp in enumerate(spws_unq):
                        # Get channel frequencies for this spw (annoyingly comes out as shape (nf, 1)
                        cfrq = spwtb.getcol('CHAN_FREQ', sp, 1)[:, 0]
                        if j == 0:
                            # Only need this the first time through
                            spwlist += [i] * len(cfrq)
                        if i == 0:
                            times[j] = tb.getcol('TIME', nbl * (i + nspw * j), 1)  # Get the time
                        spec_ = tb.getcol(datacol, nbl * (i + nspw * j), nbl)  # Get complex data for this spw
                        flag = tb.getcol('FLAG', nbl * (i + nspw * j), nbl)  # Get flags for this spw
                        nfrq = len(cfrq)
                        # Apply flags
                        if applyflag:
                            if type(fillnan) in [int, float]:
                                spec_[flag] = float(fillnan)
                            else:
                                spec_[flag] = 0.0
                        # Insert data for this spw into larger array
                        specamp[:, fptr:fptr + nfrq, :, j] = spec_
                        flagf[:, fptr:fptr + nfrq, :, j] = flag
                        freq[fptr:fptr + nfrq] = cfrq
                        fptr += nfrq
            else:
                specf = np.zeros((npol, nf, nt, nbl), complex)  # Array indexes are swapped
                # flagf = np.zeros((npol, nf, nt, nbl), int)  # Array indexes are swapped
                # ant1 = np.zeros((nt, nbl), int)  # Array indexes are swapped
                # ant2 = np.zeros((nt, nbl), int)  # Array indexes are swapped
                iptr = 0
                for j, scanid in enumerate(scanids):
                    # Loop over scans
                    s = scan_ntimes[j]
                    s1 = np.sum(scan_ntimes[:j])  # Start time index
                    s2 = np.sum(scan_ntimes[:j + 1])  # End time index
                    if verbose:
                        print('=======Filling up scan #{0:d} frm:{1:d}-{2:d}======='.format(j, s1, s2))
                    for i, sp in enumerate(spws_unq):
                        # Loop over spectral windows
                        f = spw_nfrq[i]
                        f1 = np.sum(spw_nfrq[:i])  # Start freq index
                        f2 = np.sum(spw_nfrq[:i + 1])  # End freq index
                        if verbose:
                            print('Filling up spw #{0:d} chn:{1:d}--{2:d}'.format(i, f1, f2))
                        spec_ = tb.getcol(datacol, iptr, nbl * s)
                        flag = tb.getcol('FLAG', iptr, nbl * s)
                        if j == 0:
                            cfrq = spwtb.getcol('CHAN_FREQ', sp, 1)[:, 0]
                            freq[f1:f2] = cfrq
                            spwlist += [i] * len(cfrq)
                        times[s1:s2] = tb.getcol('TIME', iptr, nbl * s).reshape(s, nbl)[:, 0]  # Get the times
                        # Apply flags
                        if applyflag:
                            if type(fillnan) in [int, float]:
                                spec_[flag] = float(fillnan)
                            else:
                                spec_[flag] = 0.0
                        # Insert data for this spw into larger array
                        specf[:, f1:f2, s1:s2] = spec_.reshape(npol, f, s, nbl)
                        # flagf[:, f1:f2, s1:s2] = flag.reshape(npol, f, s, nbl)
                        # if i==0:
                        #     ant1[s1:s2] = ant1_.reshape(s, nbl)
                        #     ant2[s1:s2] = ant2_.reshape(s, nbl)
                        # Swap the array indexes back to the desired order
                        # except:
                        #     print('error processing spw {}'.format(i))
                        iptr += nbl * s
                specamp = np.swapaxes(specf, 2, 3)
                # flagf = np.swapaxes(flagf, 2, 3)
            # if applyflag:
            #     specamp = np.ma.masked_array(specamp, flagf)
            tb.close()
            spwtb.close()
            ms.close()
            if len(antmask) > 0:
                specamp = specamp[:, :, np.where(antmask)[0], :]
            (npol, nfreq, nbl, ntim) = specamp.shape
            tim = times
            if hanning:
                os.system('rm -rf {}'.format(fname))
        else:
            # Open the ms and plot dynamic spectrum
            if verbose:
                print('Splitting selected data...')
            vis_spl = './tmpms.splitted'
            if os.path.exists(vis_spl):
                os.system('rm -rf ' + vis_spl)

            # split(vis=msfile, outputvis=vis_spl, timerange=timeran, antenna=bl, field=field, scan=scan, spw=spw,
            #       uvrange=uvrange, timebin=timebin, datacolumn=datacolumn)

            try:
                split(vis=msfile, outputvis=vis_spl, datacolumn=datacolumn, timerange=timeran, spw=spw, antenna=bl,
                      field=field, scan=scan, uvrange=uvrange, timebin=timebin)
            except:
                ms.open(msfile, nomodify=True)
                ms.split(outputms=vis_spl, whichcol=datacolumn, time=timeran, spw=spw, baseline=bl, field=field,
                         scan=scan,
                         uvrange=uvrange, timebin=timebin)
                ms.close()

            # if verbose:
            #   print('Regridding into a single spectral window...')
            # print('Reading data spw by spw')

            try:
                tb.open(vis_spl + '/POLARIZATION')
                corrtype = tb.getcell('CORR_TYPE', 0)
                pol = [stokesenum[p] for p in corrtype]
                tb.close()
            except:
                pol = []

            if regridfreq:
                if verbose:
                    print('Regridding into a single spectral window...')
                ms.open(vis_spl, nomodify=False)
                ms.cvel(outframe='LSRK', mode='frequency', interp='nearest')
                ms.selectinit(datadescid=0, reset=True)
                data = ms.getdata(['amplitude', 'time', 'axis_info'], ifraxis=True)
                specamp = data['amplitude']
                freq = data['axis_info']['freq_axis']['chan_freq']
            else:
                if verbose:
                    print('Concatenating visibility data spw by spw')
                ms.open(vis_spl)
                ms.selectinit(datadescid=0, reset=True)
                spwinfo = ms.getspectralwindowinfo()
                specamp = []
                freq = []
                time = []
                if verbose:
                    print('A total of {0:d} spws to fill'.format(len(spwinfo.keys())))
                for n, descid in enumerate(spwinfo.keys()):
                    ms.selectinit(datadescid=0, reset=True)
                    if verbose:
                        print('filling up spw #{0:d}: {1:s}'.format(n, descid))
                    descid = int(descid)
                    ms.selectinit(datadescid=n)  # , reset=True)
                    data = ms.getdata(['amplitude', 'time', 'axis_info'], ifraxis=True)
                    if verbose:
                        print('shape of this spw', data['amplitude'].shape)
                    specamp_ = data['amplitude']
                    freq_ = data['axis_info']['freq_axis']['chan_freq'].squeeze()
                    if len(freq_.shape) > 1:
                        # chan_freq for each datadecid contains the info for all the spws
                        freq = freq_.transpose().flatten()
                    else:
                        freq.append(freq_)
                    time_ = data['time']
                    if fillnan is not None:
                        flag_ = ms.getdata(['flag', 'time', 'axis_info'], ifraxis=True)['flag']
                        if type(fillnan) in [int, float]:
                            specamp_[flag_] = float(fillnan)
                        else:
                            specamp_[flag_] = 0.0
                    specamp.append(specamp_)
                    time.append(time_)
                specamp = np.concatenate(specamp, axis=1)
                try:
                    # if len(freq.shape) > 1:
                    freq = np.concatenate(freq, axis=0)
                except ValueError:
                    pass
                ms.selectinit(datadescid=0, reset=True)
            ms.close()
            if os.path.exists(vis_spl):
                os.system('rm -rf ' + vis_spl)
            (npol, nfreq, nbl, ntim) = specamp.shape
            freq = freq.reshape(nfreq)

            tim = data['time']

        if verbose:
            print('npol, nfreq, nbl, ntime:', (npol, nfreq, nbl, ntim))
        spec = np.swapaxes(specamp, 2, 1)

        if domedian:
            if verbose:
                print('doing median of all the baselines')
            # mask zero values before median
            # spec_masked = np.ma.masked_where(spec < 1e-9, spec)
            # spec_masked2 = np.ma.masked_invalid(spec)
            # spec_masked = np.ma.masked_array(spec, mask=np.logical_or(spec_masked.mask, spec_masked2.mask))
            # spec_med = np.ma.filled(np.ma.median(spec_masked, axis=1), fill_value=0.)
            spec = np.abs(spec)
            if ds_normalised == False:
                # mask zero values before median
                spec_masked = np.ma.masked_where(spec < 1e-9, spec)
                spec_masked2 = np.ma.masked_invalid(spec)
                spec_masked = np.ma.masked_array(spec, mask=np.logical_or(spec_masked.mask, spec_masked2.mask))
                spec_med = np.ma.filled(np.ma.median(spec_masked, axis=1), fill_value=0.)
                # spec_med = np.nanmedian(spec, axis=1)
                nbl = 1
                ospec = spec_med.reshape((npol, nbl, nfreq, ntim))
            else:
                spec_med_time = np.expand_dims(np.nanmedian(spec, axis=3), axis=3)
                spec_normalised = (spec - spec_med_time) / spec_med_time
                spec_med_bl = np.nanmedian(spec_normalised, axis=1)
                nbl = 1
                ospec = spec_med_bl.reshape((npol, nbl, nfreq, ntim))
                ospec = ospec * 1e4
        else:
            ospec = spec
        # Save the dynamic spectral data
        if not specfile:
            specfile = msfile + '.dspec.npz'
        if os.path.exists(specfile):
            os.system('rm -rf ' + specfile)
        np.savez(specfile, spec=ospec, tim=tim, freq=freq,
                 timeran=timeran, spw=spw, bl=bl, uvrange=uvrange, pol=pol)
        if verbose:
            print('Median dynamic spectrum saved as: ' + specfile)

        self.read(specfile)
        return specfile

    def wrt_dspec(self, specfile=None, specdat=None):
        try:
            specfile
        except NameError:
            print('No input centfile specified for reading. Aborting...')
        if not specdat:
            print('Output file name is not specified, use the default convention')
            specdat = specfile.replace('npz', 'dat')
        specdata = np.load(specfile)
        spec = np.copy(specdata['spec'][:, :, :, :])
        npl, nbl, nf, nt = spec.shape
        print('Dimension of the data cube -- # of pol, # of baseline, # of frequency, # of time:')
        print(npl, nbl, nf, nt)
        nele = npl * nbl * nf * nt
        # need to transpose to nf, nt, npl, nbl
        spec = np.transpose(spec, (2, 3, 0, 1))
        # Turn fmx into a 1d array for writing
        spec = spec.flatten()
        # Pack the data into a byte buffer.  Note the asterisks in the next three lines!
        # If your data includes integers or other data types, use 'i' or other instead of 'f'
        buf = struct.pack(str(nf) + 'f', *specdata['freq'])
        buf += struct.pack(str(nt) + 'd', *specdata['tim'])
        buf += struct.pack(str(nele) + 'f', *spec)
        with open(specdat, 'wb') as f:
            f.write(buf)
        f.close()

    def rd_dspec(self, specdata, spectype='amp', spec_unit='jy'):
        spectype = spectype.lower()
        if type(specdata) is str:
            try:
                specdata = np.load(specdata)
            except:
                specdata = np.load(specdata, encoding='latin1')
        try:
            if np.iscomplexobj(specdata['spec']):
                if spectype == 'amp':
                    if spec_unit.lower() == 'jy':
                        spec = np.abs(specdata['spec']) / 1.e4
                    elif spec_unit.lower() == 'sfu' or spec_unit.lower() == 'k':
                        spec = np.abs(specdata['spec'])
                    else:
                        raise ValueError("Input spec_unit is {}. "
                                         "If spectype = 'amp', spec_unit must be 'jy', 'sfu', or 'k'".format(spec_unit))
                elif spectype == 'pha':
                    spec = np.angle(specdata['spec'])
                else:
                    raise ValueError('spectype must be amp or phase!')
            else:
                if spec_unit.lower() == 'jy':
                    spec = specdata['spec'] / 1.e4
                elif spec_unit.lower() == 'sfu' or spec_unit.lower() == 'k':
                    spec = specdata['spec']
                else:
                    raise ValueError("Input spec_unit is {}. "
                                     "If spectype = 'amp', spec_unit must be 'jy', 'sfu', or 'k'".format(spec_unit))
            if 'bl' in specdata.keys():
                try:
                    bl = specdata['bl'].item()
                except:
                    bl = specdata['bl']
            else:
                bl = ''

            tim = specdata['tim']
            freq = specdata['freq']
            if 'pol' in specdata.keys():
                pol = specdata['pol']
            else:
                pol = []

            return spec, tim, freq, bl, pol, spec_unit
        except:
            raise ValueError('format of specdata not recognized. Check your input')

    def concat_dspec(self, specfiles, outfile=None, savespec=False):
        '''
        concatenate a list of specfiles in time axis
        :param specfiles: a list of specfile to concatenate
        :return: concatenated specdata
        '''
        from tqdm import tqdm
        if isinstance(specfiles, list):
            if len(specfiles) > 1:
                specfiles = sorted(specfiles)
            else:
                print('Abort. Only one specfile is provided.')
                return -1
        else:
            print('Please provide a list of specfiles')
            return -1

        specdata = {}

        specdata_ = np.load(specfiles[0])
        for k, v in specdata_.iteritems():
            specdata[k] = v

        for spfile in tqdm(specfiles[1:]):
            specdata_ = np.load(spfile)
            specdata['spec'] = np.concatenate((specdata['spec'], specdata_['spec']), axis=-1)
            specdata['tim'] = np.hstack((specdata['tim'], specdata_['tim']))

        if savespec:
            if not outfile:
                specfile = 'dspec.npz'
            if os.path.exists(specfile):
                os.system('rm -rf ' + specfile)
            np.savez(specfile, **specdata)
        return specdata



    def peek(self, *args, **kwargs):
        """
        Plot dynamaic spectrum onto current axes.

        Parameters
        ----------
        *args : dict

        **kwargs : dict
            Any additional plot arguments that should be used
            when plotting.

        Returns
        -------
        fig : `~matplotlib.Figure`
            A plot figure.
        """
        plt.figure()
        ret = self.plot(*args, **kwargs)
        plt.show()
        return ret

    def plot(self, pol='I', vmin=None, vmax=None, norm='log', cmap='viridis', cmap2='viridis', vmin2=None, vmax2=None, figsize=None,
             timerange=None, freqrange=None, bkgtim=None, interp_method='linear',  ignore_gaps=True, freq_unit='GHz', spec_unit=None,
             plot_fast=False, percentile=[1, 99], minmaxpercentile=False, axes=None, **kwargs):
        """
        Plots the dynamic spectrum for a given polarization.

        This method generates a plot of the dynamic spectrum, allowing for customization of the visualization through various parameters such as color maps, normalization, and value ranges.

        :param pol: Polarization for plotting. Default is 'I'.
        :type pol: str, optional
        :param vmin: Minimum intensity value for the color scale. Default is None, which auto-scales.
        :type vmin: float, optional
        :param vmax: Maximum intensity value for the color scale. Default is None, which auto-scales.
        :type vmax: float, optional
        :param norm: Normalization of the color scale. Can be 'linear', 'log', or a custom normalization. Default is 'log'.
        :type norm: str or `matplotlib.colors.Normalize`, optional
        :param cmap: Matplotlib colormap name or instance for the plot. Default is 'viridis'.
        :type cmap: str, optional
        :param cmap2: Matplotlib colormap name or instance for the second polarization plot. Default is 'viridis'.
        :type cmap2: str, optional
        :param vmin2: Minimum intensity value for the color scale of the second polarization. Default is None.
        :type vmin2: float, optional
        :param vmax2: Maximum intensity value for the color scale of the second polarization. Default is None.
        :type vmax2: float, optional
        :param timerange: Time range to plot, formatted as ['start_time', 'end_time']. Times should be in ISO format. Defaults to None, which plots the entire range.
        :type timerange: list of str, optional
        :param freqrange: Frequency range to plot, in units specified by `freq_unit`. Defaults to None, which plots the entire range.
        :type freqrange: list of float, optional
        :param bkgtim: Background time interval(s). It can be a single time interval (list of two strings)
            or multiple time intervals (list of lists of two strings each).
            Example: ['2024-06-11T19:41:00', '2024-06-11T20:40:58']
                     [['2024-06-11T19:41:00', '2024-06-11T20:40:58'], ['2024-06-11T21:41:00', '2024-06-11T22:40:58']]
        :type bkgtim: list or list of lists
        :param interp_method: The method of interpolation to use. Default is 'linear'.
            Can be any method supported by scipy.interpolate.interp1d.
        :type interp_method: str, optional
        :param ignore_gaps: If True, ignores gaps in the frequency axis. Default is True.
        :type ignore_gaps: bool, optional
        :param freq_unit: Unit for the frequency axis ('kHz', 'MHz', 'GHz'). Default is 'GHz'.
        :type freq_unit: str, optional
        :param spec_unit: Unit for the spectrum data ('sfu', 'Jy', or 'K'). If not specified, uses the unit from the Dspec object.
        :type spec_unit: str, optional
        :param plot_fast: If True, uses a faster plotting method which may reduce detail. Default is False.
        :type plot_fast: bool, optional
        :param percentile: Percentile values to use for auto-scaling the color range. Default is [1, 99].
        :type percentile: list of float, optional
        :param minmaxpercentile: If True, uses percentile for vmin and vmax. Default is False.
        :type minmaxpercentile: bool, optional
        :param kwargs: Any additional plot arguments that should be used when plotting.
        :type kwargs: dict

        :return: The matplotlib figure object containing the plot.
        :rtype: `matplotlib.figure.Figure`

        :example:
        >>> dspec = Dspec()
        >>> fig = dspec.plot(pol='I', vmin=0.1, vmax=5, cmap='hot', freqrange=[1, 2], timerange=['2021-01-01T00:00:00', '2021-01-01T01:00:00'])
        >>> plt.show()

        :notes:
        For dual-polarization plots (e.g., 'RRLL'), `cmap2` along with `vmin2` and `vmax2` can be specified to customize the appearance of the second polarization.
        """
        # Set up variables
        import matplotlib.colors as colors
        import matplotlib.pyplot as plt
        from astropy.time import Time

        if pol not in ['RR', 'LL', 'RRLL', 'XX', 'YY', 'XY', 'YX', 'XXYY', 'I', 'V', 'IV', 'IP']:
            print("Please enter 'RR', 'LL', 'RRLL','XX', 'YY', 'XY', 'YX', 'XXYY', 'I', 'V', 'IV', or 'IP' for pol")
            return 0

        spec = self.data

        try:
            cmap = copy(plt.get_cmap(cmap))
        except:
            cmap = copy(plt.get_cmap('viridis'))
        cmap.set_bad(cmap(0.0))

        if norm == 'linear':
            norm = colors.Normalize(vmax=vmax, vmin=vmin)
        if norm == 'log':
            norm = colors.LogNorm(vmax=vmax, vmin=vmin)

        if not isinstance(norm, object):
            print('color normalization is not defined as matplotlib.colors. Use default LogNorm.')
            norm = colors.LogNorm(vmax=vmax, vmin=vmin)

        bl = self.bl
        freq = self.freq_axis
        if spec_unit is None:
            spec_unit = self.spec_unit
        elif spec_unit.lower() != self.spec_unit.lower():
            if spec_unit.lower() == 'sfu' and self.spec_unit.lower() == 'jy':
                spec = np.copy(self.data) / 1e4
            elif spec_unit.lower() == 'jy' and self.spec_unit.lower() == 'sfu':
                spec = np.copy(self.data) * 1e4
            else:
                print('Spectrum unit conversion from {0:s} to {1:s} not supported'.format(self.spec_unit, spec_unit))
                print('Use the original one.')
                spec_unit = self.spec_unit
        # The following is only for the plots
        if spec_unit.lower() == 'jy':
            spec_unit_print = 'Jy'
        if spec_unit.lower() == 'k':
            spec_unit_print = 'K'
        if spec_unit.lower() == 'sfu':
            spec_unit_print = 's.f.u'

        if spec.ndim == 2:
            nfreq, ntim = len(self.freq_axis), len(self.time_axis)
            npol = 1
            nbl = 1
            polnames = self.pol
        else:
            (npol, nbl, nfreq, ntim) = spec.shape
            polnames = self.pol
            if len(polnames) != npol:
                print('The polarization dimension in the data {0:d} does not match the names {1:d}. Abort.'.format(npol,
                                                                                                                   len(polnames)))

        tim_ = self.time_axis
        tim_plt = tim_.plot_date

        if timerange:
            if isinstance(timerange[0], str):
                timerange = Time(timerange)
            tidx = np.where((tim_ >= timerange[0]) & (tim_ <= timerange[1]))[0]
        else:
            tidx = range(ntim)

        if ignore_gaps:
            df = np.median(freq[1:] - freq[:-1])
            fbreak, = np.where(freq[1:] - freq[:-1] > 2 * df)
            for n, fb in enumerate(fbreak):
                loc = fb + n + 1
                freq = np.concatenate((freq[:loc], np.array([freq[loc - 1] + df]), freq[loc:]))

        if freqrange:
            if freq_unit.lower() == 'ghz':
                fidx = np.where((freq >= freqrange[0] * 1e9) & (freq <= freqrange[1] * 1e9))[0]
            if freq_unit.lower() == 'mhz':
                fidx = np.where((freq >= freqrange[0] * 1e6) & (freq <= freqrange[1] * 1e6))[0]
            if freq_unit.lower() == 'kHz':
                fidx = np.where((freq >= freqrange[0] * 1e3) & (freq <= freqrange[1] * 1e3))[0]
        else:
            fidx = range(nfreq)

        # setup plot parameters
        print('ploting dynamic spectrum...')

        for b in range(nbl):
            if pol not in ['RRLL', 'XXYY', 'IV', 'IP']:
                if spec.ndim == 2:
                    spec_plt = spec
                else:
                    if pol in ['RR', 'XX']:
                        spec_plt = spec[0, b, :, :]
                    elif pol in ['LL', 'YY']:
                        spec_plt = spec[1, b, :, :]
                    elif pol in ['XY']:
                        spec_plt = spec[2, b, :, :]
                    elif pol in ['YX']:
                        spec_plt = spec[3, b, :, :]
                    elif pol == 'I':
                        if ('XX' in polnames) and ('YY' in polnames):
                            spec_plt = spec[0, b, :, :] + spec[1, b, :, :]
                        elif ('RR' in polnames) and ('LL' in polnames):
                            spec_plt = (spec[0, b, :, :] + spec[1, b, :, :]) / 2.
                        elif ('I' in polnames):
                            spec_plt = spec[0, b, :, :]
                    elif pol == 'V':
                        if ('XX' in polnames) and ('YY' in polnames):
                            # TODO: This does not seem to be correct. @Sijie Yu, could you please check?
                            spec_plt = spec[0, b, :, :] - spec[1, b, :, :]
                        elif ('RR' in polnames) and ('LL' in polnames):
                            spec_plt = (spec[0, b, :, :] - spec[1, b, :, :]) / 2.
                        elif ('V' in polnames):
                            spec_plt = spec[1, b, :, :]
                if ignore_gaps:
                    for n, fb in enumerate(fbreak):
                        loc = fb + n + 1
                        spec_plt = np.concatenate((spec_plt[:loc], np.zeros((1, ntim)) + np.nan, spec_plt[loc:]), 0)


                if bkgtim:
                    spec_plt -=calc_bkg_dspec(spec_plt, tim_[tidx], bkgtim, interp_method=interp_method)

                # if plot_fast:


                if figsize is None:
                    figsize = (8, 4)
                if axes:
                    ax = axes
                    fig=None
                else:
                    fig = plt.figure(figsize=figsize, dpi=100)
                    ax = fig.add_subplot(111)

                if freq_unit.lower() == 'ghz':
                    freq_plt = freq / 1e9
                if freq_unit.lower() == 'mhz':
                    freq_plt = freq / 1e6
                if freq_unit.lower() == 'khz':
                    freq_plt = freq / 1e3

                spec_plt = spec_plt[fidx, :][:, tidx]
                tim_plt = tim_plt[tidx]
                freq_plt = freq_plt[fidx]

                # Change the default for Stokes V
                if pol == 'V':
                    cmap = 'gray'
                    if (vmax is None) and (vmin is None):
                        vmax = np.nanmax(np.abs(spec_plt))
                        vmin = -vmax
                    elif (vmax is None) and not (vmin is None):
                        vmax = -vmin
                    elif not (vmax is None) and (vmin is None):
                        vmin = -vmax
                    norm = colors.Normalize(vmax=vmax, vmin=vmin)

                if minmaxpercentile:
                    if percentile[0] > 0 and percentile[1] < 100 and percentile[0] < percentile[1]:
                        norm.vmax = np.nanpercentile(spec_plt, percentile[1])
                        norm.vmin = np.nanpercentile(spec_plt, percentile[0])

                if plot_fast:
                    # rebin the data to speed up plotting
                    ds_shape = spec_plt.shape
                    if ds_shape[1] > 2100:
                        pix_rebin = ds_shape[1] // 2048
                        new_len = (ds_shape[1] // pix_rebin)
                        new_len_total = new_len * pix_rebin

                        spec_plt = rebin2d(spec_plt[:, 0:new_len_total], (ds_shape[0], new_len))
                        # tim_plt = rebin1d(tim_plt[0:new_len_total], 2048)

                    im = ax.imshow(spec_plt, cmap=cmap, norm=norm, aspect='auto', origin='lower',
                                   extent=[tim_plt[0], tim_plt[-1], freq_plt[0], freq_plt[-1]])

                else:
                    im = ax.pcolormesh(tim_plt, freq_plt, spec_plt, cmap=cmap, norm=norm, shading='auto',
                                       rasterized=True)
                    ax.set_xlim(tim_plt[0], tim_plt[-1])
                    ax.set_ylim(freq_plt[0], freq_plt[-1])

                    def format_coord(x, y):
                        col = np.argmin(np.absolute(tim_plt - x))
                        row = np.argmin(np.absolute(freq_plt - y))
                        if col >= 0 and col < ntim and row >= 0 and row < nfreq:
                            timstr = tim_[col].isot
                            flux = spec_plt[row, col]
                            return 'time {0} = {1}, freq {2} = {3:.3f} {4:s}, flux = {5:.2f} {6:s}'.format(col, timstr,
                                                                                                           row,
                                                                                                           y, freq_unit,
                                                                                                           flux,
                                                                                                           spec_unit_print)
                        else:
                            return 'x = {0}, y = {1:.3f}'.format(x, y)

                    ax.format_coord = format_coord

                # make colorbar
                divider = make_axes_locatable(ax)
                cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
                clb_spec = plt.colorbar(im, ax=ax, cax=cax_spec)
                clb_spec.set_label('Intensity [{}]'.format(spec_unit_print))


                ax.set_ylabel('Frequency [{0:s}]'.format(freq_unit))
                if bl:
                    ax.set_title('{0:s}-{1:s} dynamic spectrum @ bl {2:s} for pol {3:s}'.
                                 format(self.observatory, self.telescope, bl.split(';')[b], pol))
                else:
                    ax.set_title('{0:s}-{1:s} dynamic spectrum for pol {2:s}'.
                                 format(self.observatory, self.telescope, pol))
                locator = AutoDateLocator(minticks=2)
                ax.xaxis.set_major_locator(locator)
                formatter = AutoDateFormatter(locator)
                formatter.scaled[1 / 24] = '%D %H'
                formatter.scaled[1 / (24 * 60)] = '%H:%M'
                ax.xaxis.set_major_formatter(formatter)
                ax.set_autoscale_on(False)

            else:
                if figsize is None:
                    figsize = (8, 6)
                fig = plt.figure(figsize=figsize, dpi=100)
                if ('RR' in polnames) and ('LL' in polnames):
                    R_plot = np.absolute(spec[0, b, :, :])
                    L_plot = np.absolute(spec[1, b, :, :])
                    I_plot = (R_plot + L_plot) / 2.
                    V_plot = (R_plot - L_plot) / 2.
                if ('I' in polnames) and ('V' in polnames or 'P' in polnames) :
                    I_plot = spec[0, b, :, :]
                    V_plot = spec[1, b, :, :]
                    R_plot = I_plot + V_plot
                    L_plot = I_plot - V_plot
                if pol in ['RRLL', 'XXYY']:
                    spec_plt_1 = R_plot
                    spec_plt_2 = L_plot
                    polstr = [pol[:2], pol[2:]]
                elif pol == 'IV':
                    spec_plt_1 = I_plot
                    spec_plt_2 = V_plot
                    cmap2 = 'gray'
                    if (vmax2 is None) and (vmin2 is None):
                        vmax2 = np.nanmax(np.abs(spec_plt_2))
                        vmin2 = -vmax2
                    elif (vmax2 is None) and not (vmin2 is None):
                        vmax2 = -vmin2
                    elif not (vmax2 is None) and (vmin2 is None):
                        vmin2 = -vmax2
                    polstr = ['I', 'V']
                elif pol == 'IP':
                    # this is for Stokes I + polarization degree
                    spec_plt_1 = I_plot
                    spec_plt_2 = V_plot / I_plot
                    cmap2 = 'gray'
                    if (vmax2 is None) and (vmin2 is None):
                        vmax2 = 1.
                        vmin2 = -1.
                    elif (vmax2 is None) and not (vmin2 is None):
                        vmax2 = -vmin2
                    elif not (vmax2 is None) and (vmin2 is None):
                        vmin2 = -vmax2
                    polstr = ['I', 'V/I']

                if ignore_gaps:
                    for n, fb in enumerate(fbreak):
                        loc = fb + n + 1
                        spec_plt_1 = np.concatenate((spec_plt_1[:loc], np.zeros((1, ntim)) + np.nan, spec_plt_1[loc:]),
                                                    0)
                        spec_plt_2 = np.concatenate((spec_plt_2[:loc], np.zeros((1, ntim)) + np.nan, spec_plt_2[loc:]),
                                                    0)


                if bkgtim:
                    spec_plt_1 -= calc_bkg_dspec(spec_plt_1, tim_[tidx], bkgtim, interp_method=interp_method)
                    spec_plt_2 -= calc_bkg_dspec(spec_plt_2, tim_[tidx], bkgtim, interp_method=interp_method)

                ax1 = fig.add_subplot(211)
                if freq_unit.lower() == 'ghz':
                    freq_plt = freq / 1e9
                if freq_unit.lower() == 'mhz':
                    freq_plt = freq / 1e6
                if freq_unit.lower() == 'khz':
                    freq_plt = freq / 1e3

                spec_plt_1 = spec_plt_1[fidx, :][:, tidx]
                spec_plt_2 = spec_plt_2[fidx, :][:, tidx]
                tim_plt = tim_plt[tidx]
                freq_plt = freq_plt[fidx]

                if minmaxpercentile:
                    if percentile[0] > 0 and percentile[1] < 100 and percentile[0] < percentile[1]:
                        norm.vmax = np.nanpercentile(spec_plt_1, percentile[1])
                        norm.vmin = np.nanpercentile(spec_plt_1, percentile[0])

                if plot_fast:
                    # compress in time (idx1)
                    ds_shape = spec_plt_1.shape
                    if ds_shape[1] > 2100:
                        pix_rebin = ds_shape[1] // 2048
                        new_len = (ds_shape[1] // pix_rebin)
                        new_len_total = new_len * pix_rebin

                        spec_plt_1 = rebin2d(spec_plt_1[:, 0:new_len_total], (ds_shape[0], new_len))
                        spec_plt_2 = rebin2d(spec_plt_2[:, 0:new_len_total], (ds_shape[0], new_len))
                        # tim_plt = rebin1d(tim_plt[0:new_len_total], new_len)

                    im = ax1.imshow(spec_plt_1, cmap=cmap, norm=norm, aspect='auto', origin='lower',
                                    extent=[tim_plt[0], tim_plt[-1], freq_plt[0], freq_plt[-1]])
                else:
                    im = ax1.pcolormesh(tim_plt, freq_plt, spec_plt_1, cmap=cmap, norm=norm, shading='auto',
                                        rasterized=True)
                    ax1.set_xlim(tim_plt[0], tim_plt[-1])
                    ax1.set_ylim(freq_plt[0], freq_plt[-1])

                    def format_coord(x, y):
                        col = np.argmin(np.absolute(tim_plt - x))
                        row = np.argmin(np.absolute(freq_plt - y))
                        if col >= 0 and col < ntim and row >= 0 and row < nfreq:
                            timstr = tim_[col].isot
                            flux = spec_plt_1[row, col]
                            return 'time {0} = {1}, freq {2} = {3:.3f} {4:s}, value = {5:.2f} {6:s}'.format(col, timstr,
                                                                                                            row,
                                                                                                            y,
                                                                                                            freq_unit,
                                                                                                            flux,
                                                                                                            spec_unit_print)
                        else:
                            return 'x = {0}, y = {1:.3f}'.format(x, y)

                    ax1.format_coord = format_coord

                divider = make_axes_locatable(ax1)
                cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
                clb_spec = plt.colorbar(im, ax=ax1, cax=cax_spec)
                clb_spec.set_label('Intensity [{}]'.format(spec_unit_print))


                ax1.set_ylabel('Frequency [{0:s}]'.format(freq_unit))

                locator = AutoDateLocator(minticks=2)
                ax1.xaxis.set_major_locator(locator)
                # ax1.xaxis.set_major_formatter(AutoDateFormatter(locator))
                formatter = AutoDateFormatter(locator)
                formatter.scaled[1 / 24] = '%D %H'
                formatter.scaled[1 / (24 * 60)] = '%H:%M'
                ax1.xaxis.set_major_formatter(formatter)
                if bl:
                    ax1.set_title('{0:s}-{1:s} dynamic spectrum @ bl {2:s} for pol {3:s}'. \
                                  format(self.observatory, self.telescope, bl.split(';')[b], polstr[0]))
                else:
                    ax1.set_title('{0:s}-{1:s} dynamic spectrum for pol {2:s}'. \
                                  format(self.observatory, self.telescope, polstr[0]))

                ax1.set_autoscale_on(False)

                ax2 = fig.add_subplot(212, sharex=ax1, sharey=ax1)
                if cmap2 is None:
                    cmap2 = cmap
                if vmin2 is None:
                    vmin2 = vmin
                if vmax2 is None:
                    vmax2 = vmax
                norm2 = colors.Normalize(vmax=vmax2, vmin=vmin2)



                if plot_fast:
                    im = ax2.imshow(spec_plt_2, cmap=cmap2, norm=norm2, aspect='auto', origin='lower',
                                    extent=[tim_plt[0], tim_plt[-1], freq_plt[0], freq_plt[-1]])
                else:
                    im = ax2.pcolormesh(tim_plt, freq_plt, spec_plt_2, cmap=cmap2, norm=norm2, shading='auto',
                                        rasterized=True)
                    def format_coord(x, y):
                        col = np.argmin(np.absolute(tim_plt - x))
                        row = np.argmin(np.absolute(freq_plt - y))
                        if col >= 0 and col < ntim and row >= 0 and row < nfreq:
                            timstr = tim_[col].isot
                            flux = spec_plt_2[row, col]
                            return 'time {0} = {1}, freq {2} = {3:.3f} {4:s}, value = {5:.2f} {6:s}'.format(col, timstr,
                                                                                                            row,
                                                                                                            y,
                                                                                                            freq_unit,
                                                                                                            flux,
                                                                                                            spec_unit_print)
                        else:
                            return 'x = {0}, y = {1:.3f}'.format(x, y)
                    ax2.format_coord = format_coord

                    ax2.set_xlim(tim_plt[0], tim_plt[-1])
                    ax2.set_ylim(freq_plt[0], freq_plt[-1])

                divider = make_axes_locatable(ax2)
                cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
                clb_spec = plt.colorbar(im, ax=ax2, cax=cax_spec)
                if pol == 'IP':
                    clb_spec.set_label(polstr[1])
                else:
                    clb_spec.set_label('Intensity [{}]'.format(spec_unit_print))

                locator = AutoDateLocator(minticks=2)
                ax2.xaxis.set_major_locator(locator)
                formatter = AutoDateFormatter(locator)
                formatter.scaled[1 / 24] = '%D %H'
                formatter.scaled[1 / (24 * 60)] = '%H:%M'
                ax2.xaxis.set_major_formatter(formatter)

                ax2.set_ylabel('Frequency [{0:s}]'.format(freq_unit))
                if bl:
                    ax2.set_title('{0:s}-{1:s} dynamic spectrum @ bl {2:s} for pol {3:s}'. \
                                  format(self.observatory, self.telescope, bl.split(';')[b], polstr[1]))
                else:
                    ax2.set_title('{0:s}-{1:s} dynamic spectrum for pol {2:s}'. \
                                  format(self.observatory, self.telescope, polstr[1]))

                ax2.set_autoscale_on(False)

            if fig:
                fig.tight_layout()
        return fig



import numpy as np

sin = np.sin
cos = np.cos


def hgs2hcc(rsun, lon, lat, B0, L0):
    lon_L0 = lon - L0
    x = rsun * cos(lat) * sin(lon)
    y = rsun * (sin(lat) * cos(B0) - cos(lat) * cos(lon_L0) * sin(B0))
    z = rsun * (sin(lat) * sin(B0) + cos(lat) * cos(lon_L0) * cos(B0))
    return x, y, z


def hcc2hgs(x, y, z, B0, L0):
    rsun = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    lat = np.arcsin((y * cos(B0) + z * sin(B0)) / rsun)
    lon = L0 + np.arctan2(x, z * cos(B0) - y * sin(B0))
    return rsun, lon, lat



##################### generated by xml-casa (v2) from importeovsa.xml ###############
##################### c9df476ec6913d95cc0ec94dbdf76dda ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from suncasatasks import importeovsa as _importeovsa_t
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _importeovsa:
    """
    importeovsa ---- Parallelized import EOVSA idb file(s) to a measurement set or multiple measurement set.

    
    Parallelized imports an arbitrary number of EOVSA idb-format data sets into
    a casa measurement set. If more than one band is present, they
    will be put in the same measurement set but in a separate spectral
    window.
    

    --------- parameter descriptions ---------------------------------------------

    idbfiles          Name of input EOVSA idb file(s) or observation time range.
    ncpu              Number of cpu cores to use
    timebin           Bin width for time averaging
    width             Width of output channel relative to MS channel (# to average)
    visprefix         Prefix of vis names (may include the path).
    udb_corr          if applying correction to input UDB files before import to MS.
    nocreatms         If setting nocreatms True, will simulate a model measurement set for the first idb file and copy the model for the rest of idl files in list. If False, will simulate a new measurement set for every idbfile in list.
    doconcat          If concatenate multi casa measurement sets to one file.
    modelms           Name of input model measurement set file. If modelms is assigned, no simulation will start.
    doscaling         If creating a new MS file with the amplitude of visibility data rescaled.
    keep_nsclms       Keep the no scaling measurement sets
    use_exist_udbcorr If use the existed udb_corr results.

    --------- examples -----------------------------------------------------------

    
    
    Parallelized imports an arbitrary number of EOVSA idb-format data sets into
    a casa measurement set. If more than one band is present, they
    will be put in the same measurement set but in a separate spectral
    window.
    
    Detailed Keyword arguments:
    
    idbfiles -- Name of input EOVSA idb file(s)
    default: none. Must be supplied
    example: idbfiles = 'IDB20160524000518'
    example: idbfiles=['IDB20160524000518','IDB20160524000528']
    
    ncpu -- Number of cpu cores to use
    default: 8
    
    visprefix -- Prefix of vis names (may include the path)
    default: none;
    example: visprefix='sun/']
    
    
    --- Data Selection ---
    
    nocreatms -- If copying a new MS file instead of create one from MS simulator.
    default: False
    
    modelms -- Name of the standard Measurement Set. IF modelms is not provided, use
    '/home/user/sjyu/20160531/ms/sun/SUN/SUN_20160531T142234-10m.1s.ms' as a standard MS.
    
    doconcat -- If outputing one single MS file
    
    
    --- Channel averaging parameter ---
    
    width -- Number of input channels to average to create an output
    channel. If a list is given, each bin will apply to one spw in
    the selection.
    default: 1 => no channel averaging.
    options: (int) or [int]
    
    example: chanbin=[2,3] => average 2 channels of 1st selected
    spectral window and 3 in the second one.
    
    
    --- Time averaging parameters ---
    
    timebin -- Bin width for time averaging. When timebin is greater than 0s,
    the task will average data in time. Flagged data will be included
    in the average calculation, unless the parameter keepflags is set to False.
    In this case only partially flagged rows will be used in the average.
    default: '0s'
    
    
    
    
    
    
    
    
    
    


    """

    _info_group_ = """Import/export"""
    _info_desc_ = """Parallelized import EOVSA idb file(s) to a measurement set or multiple measurement set."""

    __schema = {'idbfiles': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'ncpu': {'type': 'cInt'}, 'timebin': {'type': 'cStr', 'coerce': _coerce.to_str}, 'width': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, {'type': 'cInt'}, {'type': 'cIntVec', 'coerce': [_coerce.to_list,_coerce.to_intvec]}]}, 'visprefix': {'type': 'cStr', 'coerce': _coerce.to_str}, 'udb_corr': {'type': 'cBool'}, 'nocreatms': {'type': 'cBool'}, 'doconcat': {'type': 'cBool'}, 'modelms': {'type': 'cStr', 'coerce': _coerce.to_str}, 'doscaling': {'type': 'cBool'}, 'keep_nsclms': {'type': 'cBool'}, 'use_exist_udbcorr': {'type': 'cBool'}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 14 + 4
        output = [ ]
        addon = ''
        first_addon = True
        if len(description) == 0:
            out.write(param_prefix + " #\n")
            return
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __width_dflt( self, glb ):
        return int(1)

    def __width( self, glb ):
        if 'width' in glb: return glb['width']
        return int(1)

    def __idbfiles_dflt( self, glb ):
        return ''

    def __idbfiles( self, glb ):
        if 'idbfiles' in glb: return glb['idbfiles']
        return ''

    def __doscaling_dflt( self, glb ):
        return False

    def __doscaling( self, glb ):
        if 'doscaling' in glb: return glb['doscaling']
        return False

    def __ncpu_dflt( self, glb ):
        return int(1)

    def __ncpu( self, glb ):
        if 'ncpu' in glb: return glb['ncpu']
        return int(1)

    def __modelms_dflt( self, glb ):
        return ''

    def __modelms( self, glb ):
        if 'modelms' in glb: return glb['modelms']
        return ''

    def __visprefix_dflt( self, glb ):
        return ''

    def __visprefix( self, glb ):
        if 'visprefix' in glb: return glb['visprefix']
        return ''

    def __timebin_dflt( self, glb ):
        return '0s'

    def __timebin( self, glb ):
        if 'timebin' in glb: return glb['timebin']
        return '0s'

    def __udb_corr_dflt( self, glb ):
        return True

    def __udb_corr( self, glb ):
        if 'udb_corr' in glb: return glb['udb_corr']
        return True

    def __nocreatms_dflt( self, glb ):
        return False

    def __nocreatms( self, glb ):
        if 'nocreatms' in glb: return glb['nocreatms']
        return False

    def __doconcat_dflt( self, glb ):
        return False

    def __doconcat( self, glb ):
        if 'doconcat' in glb: return glb['doconcat']
        return False

    #--------- return non subparam/when values ---------------------------------------------
    def __use_exist_udbcorr( self, glb ):
        if 'use_exist_udbcorr' in glb: return glb['use_exist_udbcorr']
        return False

    #--------- return inp/go default --------------------------------------------------
    def __use_exist_udbcorr_dflt( self, glb ):
        if self.__udb_corr( glb ) == bool(True): return bool(False)
        return None
    def __keep_nsclms_dflt( self, glb ):
        if self.__doscaling( glb ) == bool(True): return bool(False)
        return None

    #--------- return subparam values -------------------------------------------------
    def __keep_nsclms( self, glb ):
        if 'keep_nsclms' in glb: return glb['keep_nsclms']
        dflt = self.__keep_nsclms_dflt( glb )
        if dflt is not None: return dflt
        return False

    #--------- subparam inp output ----------------------------------------------------
    def __idbfiles_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__idbfiles_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = ''
        value = self.__idbfiles( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'idbfiles': value},{'idbfiles': self.__schema['idbfiles']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('idbfiles',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __ncpu_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__ncpu_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return int(1)
        description = ''
        value = self.__ncpu( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'ncpu': value},{'ncpu': self.__schema['ncpu']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('ncpu',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __timebin_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__timebin_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return '0s'
        description = ''
        value = self.__timebin( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'timebin': value},{'timebin': self.__schema['timebin']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('timebin',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __width_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__width_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return int(1)
        description = ''
        value = self.__width( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'width': value},{'width': self.__schema['width']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('width',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __visprefix_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__visprefix_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = ''
        value = self.__visprefix( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'visprefix': value},{'visprefix': self.__schema['visprefix']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('visprefix',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __udb_corr_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__udb_corr_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        description = ''
        value = self.__udb_corr( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'udb_corr': value},{'udb_corr': self.__schema['udb_corr']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-14.14s =\x1B[0m %s%-23s%s' % ('udb_corr',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __nocreatms_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__nocreatms_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__nocreatms( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'nocreatms': value},{'nocreatms': self.__schema['nocreatms']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('nocreatms',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __doconcat_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doconcat_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__doconcat( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doconcat': value},{'doconcat': self.__schema['doconcat']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('doconcat',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __modelms_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__modelms_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = ''
        value = self.__modelms( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'modelms': value},{'modelms': self.__schema['modelms']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('modelms',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __doscaling_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doscaling_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__doscaling( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doscaling': value},{'doscaling': self.__schema['doscaling']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-14.14s =\x1B[0m %s%-23s%s' % ('doscaling',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __keep_nsclms_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__keep_nsclms_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        if self.__keep_nsclms_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__keep_nsclms( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'keep_nsclms': value},{'keep_nsclms': self.__schema['keep_nsclms']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-11.11s =\x1B[0m %s%-23s%s' % ('keep_nsclms',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __use_exist_udbcorr_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__use_exist_udbcorr_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__use_exist_udbcorr( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'use_exist_udbcorr': value},{'use_exist_udbcorr': self.__schema['use_exist_udbcorr']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-14.14s = %s%-23s%s' % ('use_exist_udbcorr',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'doscaling' in glb: del glb['doscaling']
        if 'use_exist_udbcorr' in glb: del glb['use_exist_udbcorr']
        if 'idbfiles' in glb: del glb['idbfiles']
        if 'nocreatms' in glb: del glb['nocreatms']
        if 'timebin' in glb: del glb['timebin']
        if 'modelms' in glb: del glb['modelms']
        if 'visprefix' in glb: del glb['visprefix']
        if 'doconcat' in glb: del glb['doconcat']
        if 'ncpu' in glb: del glb['ncpu']
        if 'udb_corr' in glb: del glb['udb_corr']
        if 'keep_nsclms' in glb: del glb['keep_nsclms']
        if 'width' in glb: del glb['width']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# importeovsa -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__idbfiles_inp( )
        self.__ncpu_inp( )
        self.__timebin_inp( )
        self.__width_inp( )
        self.__visprefix_inp( )
        self.__udb_corr_inp( )
        self.__nocreatms_inp( )
        self.__doconcat_inp( )
        self.__modelms_inp( )
        self.__doscaling_inp( )
        self.__keep_nsclms_inp( )
        self.__use_exist_udbcorr_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,savefile=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = savefile
        if filename is None:
            filename = "importeovsa.last" if os.path.isfile("importeovsa.last") else "importeovsa.saved"
        if os.path.isfile(filename):
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file: %s\nsetting defaults instead..." % filename)
            self.set_global_defaults( )

    #--------- tput function ----------------------------------------------------------
    def tput(self,outfile=None):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s

        _postfile = outfile if outfile is not None else os.path.realpath('importeovsa.last')

        _invocation_parameters = OrderedDict( )
        _invocation_parameters['idbfiles'] = self.__idbfiles( self.__globals_( ) )
        _invocation_parameters['ncpu'] = self.__ncpu( self.__globals_( ) )
        _invocation_parameters['timebin'] = self.__timebin( self.__globals_( ) )
        _invocation_parameters['width'] = self.__width( self.__globals_( ) )
        _invocation_parameters['visprefix'] = self.__visprefix( self.__globals_( ) )
        _invocation_parameters['udb_corr'] = self.__udb_corr( self.__globals_( ) )
        _invocation_parameters['nocreatms'] = self.__nocreatms( self.__globals_( ) )
        _invocation_parameters['doconcat'] = self.__doconcat( self.__globals_( ) )
        _invocation_parameters['modelms'] = self.__modelms( self.__globals_( ) )
        _invocation_parameters['doscaling'] = self.__doscaling( self.__globals_( ) )
        _invocation_parameters['keep_nsclms'] = self.__keep_nsclms( self.__globals_( ) )
        _invocation_parameters['use_exist_udbcorr'] = self.__use_exist_udbcorr( self.__globals_( ) )

        try:
            with open(_postfile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-11s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#importeovsa( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: return False
        return True

    def __call__( self, idbfiles=None, ncpu=None, timebin=None, width=None, visprefix=None, udb_corr=None, nocreatms=None, doconcat=None, modelms=None, doscaling=None, keep_nsclms=None, use_exist_udbcorr=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('importeovsa.pre')
        _postfile = os.path.realpath('importeovsa.last')
        _return_result_ = None
        _arguments = [idbfiles,ncpu,timebin,width,visprefix,udb_corr,nocreatms,doconcat,modelms,doscaling,keep_nsclms,use_exist_udbcorr]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if idbfiles is not None: local_global['idbfiles'] = idbfiles
            if ncpu is not None: local_global['ncpu'] = ncpu
            if timebin is not None: local_global['timebin'] = timebin
            if width is not None: local_global['width'] = width
            if visprefix is not None: local_global['visprefix'] = visprefix
            if udb_corr is not None: local_global['udb_corr'] = udb_corr
            if nocreatms is not None: local_global['nocreatms'] = nocreatms
            if doconcat is not None: local_global['doconcat'] = doconcat
            if modelms is not None: local_global['modelms'] = modelms
            if doscaling is not None: local_global['doscaling'] = doscaling
            if use_exist_udbcorr is not None: local_global['use_exist_udbcorr'] = use_exist_udbcorr

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['idbfiles'] = self.__idbfiles( local_global )
            _invocation_parameters['ncpu'] = self.__ncpu( local_global )
            _invocation_parameters['timebin'] = self.__timebin( local_global )
            _invocation_parameters['width'] = self.__width( local_global )
            _invocation_parameters['visprefix'] = self.__visprefix( local_global )
            _invocation_parameters['udb_corr'] = self.__udb_corr( local_global )
            _invocation_parameters['nocreatms'] = self.__nocreatms( local_global )
            _invocation_parameters['doconcat'] = self.__doconcat( local_global )
            _invocation_parameters['modelms'] = self.__modelms( local_global )
            _invocation_parameters['doscaling'] = self.__doscaling( local_global )
            _invocation_parameters['use_exist_udbcorr'] = self.__use_exist_udbcorr( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            _invocation_parameters['keep_nsclms'] = self.__keep_nsclms( _invocation_parameters ) if keep_nsclms is None else keep_nsclms

        else:
            # invoke with inp/go semantics
            _invocation_parameters['idbfiles'] = self.__idbfiles( self.__globals_( ) )
            _invocation_parameters['ncpu'] = self.__ncpu( self.__globals_( ) )
            _invocation_parameters['timebin'] = self.__timebin( self.__globals_( ) )
            _invocation_parameters['width'] = self.__width( self.__globals_( ) )
            _invocation_parameters['visprefix'] = self.__visprefix( self.__globals_( ) )
            _invocation_parameters['udb_corr'] = self.__udb_corr( self.__globals_( ) )
            _invocation_parameters['nocreatms'] = self.__nocreatms( self.__globals_( ) )
            _invocation_parameters['doconcat'] = self.__doconcat( self.__globals_( ) )
            _invocation_parameters['modelms'] = self.__modelms( self.__globals_( ) )
            _invocation_parameters['doscaling'] = self.__doscaling( self.__globals_( ) )
            _invocation_parameters['keep_nsclms'] = self.__keep_nsclms( self.__globals_( ) )
            _invocation_parameters['use_exist_udbcorr'] = self.__use_exist_udbcorr( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-11s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#importeovsa( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _return_result_ = _importeovsa_t( _invocation_parameters['idbfiles'],_invocation_parameters['ncpu'],_invocation_parameters['timebin'],_invocation_parameters['width'],_invocation_parameters['visprefix'],_invocation_parameters['udb_corr'],_invocation_parameters['nocreatms'],_invocation_parameters['doconcat'],_invocation_parameters['modelms'],_invocation_parameters['doscaling'],_invocation_parameters['keep_nsclms'],_invocation_parameters['use_exist_udbcorr'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('importeovsa')
            casalog.post("Exception Reported: Error in importeovsa: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            _return_result_ = False
        try:
            os.rename(_prefile,_postfile)
        except: pass
        return _return_result_

importeovsa = _importeovsa( )

